#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <gtk/gtk.h>
#include "userfuncts.h"
#include "list.h"
#include <stdio.h>

#define LANCZOS     1
#define TRACEMIN    2
#define DAVIDSON    3
#define ARNOLDI     4

#define SINPI_3    0.8660254
#define COSPI_3    0.5  
#define Zi_CONV            0x01
#define Ze_CONV            0x02
#define Zl_INCL            0x04
#define Zl_CONV            0x08

#define SOCK_BUF_SIZE     500

extern GtkRuler *HRuler,*VRuler;
extern GtkDrawingArea *LevelDrawingArea,*MachineDrawingArea;
extern GdkPixmap *LevelPixmap,*MachinePixmap;
static GdkGC *LevelGC=NULL,*MachinesGC=NULL;
static int Machines=0;
/* function takes a string representing a complex
   and returns real and imaginary real part */

int ProcessComplex(char *s,float *RePtr,float *ImPtr)
{
  if (sscanf(s,"%e %ei\n",RePtr,ImPtr)==2)
    return 2;
  if (sscanf(s,"%e + %ei\n",RePtr,ImPtr)==2)
    return 2;
  if (sscanf(s,"%e - %ei\n",RePtr,ImPtr)==2)
    {
      *ImPtr=- *ImPtr;
      return 2;
    }
  if (sscanf(s,"%e",RePtr)==1)
    {
      *ImPtr=0;
      return 1;
    }
  return -1;
}

/* Socket Functions */
static char HostName[100]="localhost.localdomain";
static char Port[10]="10000";
static int SockNum=-1;

char *GetCurHostName()
{
  return HostName;
}

char *GetCurPort()
{
  return Port;
}

int SetHostName(char *s)
{
  strncpy(HostName,s,100);
}

int SetHostPort(char *s)
{
  strncpy(Port,s,10);
}


/* Connecting and disconnecting MainMaster */
struct padcmd
{
  int fd;
  char string[60];
  gpointer data;
};


static struct padcmd cmdinfo;



int Connect2MainMaster()
{
  struct sockaddr_in nom;
  struct hostent *hp;
  extern int file_function(struct padcmd *cmdinfo);

  if ((SockNum=socket(AF_INET,SOCK_STREAM,0))==-1)
    {
      perror("Could not create socket");
      return 0;
    }
  if ((hp=gethostbyname(HostName))==NULL)
    {
      printf("%s host unknown\n",HostName);
      return 0;
    }
  bcopy(hp->h_addr,&nom.sin_addr,hp->h_length);
  nom.sin_family=AF_INET;
  nom.sin_port=htons(atoi(Port));
  if (connect(SockNum,&nom,sizeof(nom))==-1)
    {
      perror("connect");
      return 0;
    }
  cmdinfo.fd=SockNum;
  gdk_input_add(SockNum,GDK_INPUT_READ,GTK_SIGNAL_FUNC(file_function),&cmdinfo);
}

int DisconnectFromMainMaster()
{
  if (SockNum!=-1)
    {
      char s[]="Q";
      printf("Sending \"Q\" to socket...\n");fflush(stdout);
      write(SockNum,s,2);
      close(SockNum);
    }
}


/* Handles Cluster messages */

struct Segment
{
  long xi,yi,xe,ye;
  char Flag;
  double x,y;
  double Sigma;
};




GtkWidget* ErrorMsg(char *msg)
{
  GtkWidget *dialog3;
  GtkWidget *dialog_vbox3;
  GtkWidget *hbox2;
  GtkWidget *pixmap7;
  GtkWidget *label72;
  GtkWidget *dialog_action_area3;
  GtkWidget *button11;
  GtkWidget*  create_pixmap (GtkWidget       *widget,
			     const gchar     *filename);

  dialog3 = gtk_dialog_new ();
  gtk_widget_set_name (dialog3, "dialog3");
  gtk_object_set_data (GTK_OBJECT (dialog3), "dialog3", dialog3);
  gtk_window_set_title (GTK_WINDOW (dialog3), "Error");
  gtk_window_set_policy (GTK_WINDOW (dialog3), TRUE, TRUE, FALSE);

  dialog_vbox3 = GTK_DIALOG (dialog3)->vbox;
  gtk_widget_set_name (dialog_vbox3, "dialog_vbox3");
  gtk_object_set_data (GTK_OBJECT (dialog3), "dialog_vbox3", dialog_vbox3);
  gtk_widget_show (dialog_vbox3);

  hbox2 = gtk_hbox_new (FALSE, 15);
  gtk_widget_set_name (hbox2, "hbox2");
  gtk_widget_ref (hbox2);
  gtk_object_set_data_full (GTK_OBJECT (dialog3), "hbox2", hbox2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox2);
  gtk_box_pack_start (GTK_BOX (dialog_vbox3), hbox2, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (hbox2), 17);

  pixmap7 = create_pixmap (dialog3, "Error.xpm");
  gtk_widget_set_name (pixmap7, "pixmap7");
  gtk_widget_ref (pixmap7);
  gtk_object_set_data_full (GTK_OBJECT (dialog3), "pixmap7", pixmap7,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (pixmap7);
  gtk_box_pack_start (GTK_BOX (hbox2), pixmap7, FALSE, FALSE, 0);

  label72 = gtk_label_new (msg);
  gtk_widget_set_name (label72, "label72");
  gtk_widget_ref (label72);
  gtk_object_set_data_full (GTK_OBJECT (dialog3), "label72", label72,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label72);
  gtk_box_pack_start (GTK_BOX (hbox2), label72, FALSE, FALSE, 0);

  dialog_action_area3 = GTK_DIALOG (dialog3)->action_area;
  gtk_widget_set_name (dialog_action_area3, "dialog_action_area3");
  gtk_object_set_data (GTK_OBJECT (dialog3), "dialog_action_area3", dialog_action_area3);
  gtk_widget_show (dialog_action_area3);
  gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area3), 10);

  button11 = gtk_button_new_with_label ("OK");
  gtk_widget_set_name (button11, "button11");
  gtk_widget_ref (button11);
  gtk_object_set_data_full (GTK_OBJECT (dialog3), "button11", button11,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (button11);
  gtk_box_pack_start (GTK_BOX (dialog_action_area3), button11, FALSE, FALSE, 0);
  gtk_widget_set_usize (button11, 60, 30);

  gtk_signal_connect_object (GTK_OBJECT (button11), "clicked",
                             GTK_SIGNAL_FUNC (gtk_widget_destroy),
                             GTK_OBJECT (dialog3));

  return dialog3;
}



int file_function(struct padcmd *cmdinfo)
{
  static char Buffer[SOCK_BUF_SIZE];
  static char s[1000];
  char *msg;
  int n;
  static int k=0;
  struct Segment Seg;

  n=read(SockNum,Buffer+k,SOCK_BUF_SIZE-k);
  if (n==SOCK_BUF_SIZE-k) /* Too many incoming bytes */
    {
      n=SOCK_BUF_SIZE-1;
      while(Buffer[n]!='\n')
	n--;
      Buffer[n]='\0';
      k=SOCK_BUF_SIZE-1-n;
    }
  else
    {
      Buffer[n+k-1]='\0';
      k=0;
    }
  msg=strtok(Buffer,"\n");
  while(msg)
    {
      switch(msg[0])
	{
	
	case 'X':
	  {
	    GtkWidget *w;
	    w=ErrorMsg(&msg[2]);
	    gtk_widget_show(w);
	    break;
	  }
	case 'B':
	  {
	    double r,i,f;
	    long It;
	    GtkWidget *window;
	    GtkWidget *Label;
	    sscanf(msg,"B %le %le %le %ld",&r,&i,&f,&It);
	    window = gtk_window_new (GTK_WINDOW_DIALOG);
	    gtk_window_set_title(GTK_WINDOW(window),"Return Value"); 
	    sprintf(s,"S(%e%+ei)=%e   (%ld Iterations)",r,i,f,It);
	    Label=gtk_label_new(s);
	    gtk_container_add(GTK_CONTAINER(window),Label);
	    gtk_widget_show(Label);
	    gtk_widget_show(window);	
	    break;
	  }
	case 'A':
	  {
	    long Id,Flag,Method,SigmaMaxIt;
	    double Tau,Sigma,SigmaPrec,SplitPrec,CreatedTime;


	    sscanf(msg,"A %ld %ld %ld %le %le %le %le %ld %lf",
		   &Id,&Flag,&Method,
		   &Tau,&Sigma,&SigmaPrec,&SplitPrec,
		   &SigmaMaxIt,&CreatedTime);
	    CreateGLevelCurve(Id,(char) Flag,(char) Method,Tau,
			      Sigma,SigmaPrec,SplitPrec,
			      SigmaMaxIt,CreatedTime,(GtkWidget *)LevelDrawingArea);
	    break;
	  }
	case 'D':
	  {
	    int Status;
	    long Id;

	    if (sscanf(msg,"D %ld %ld %ld %ld %ld %d",
		   &Id,&Seg.xi,&Seg.yi,&Seg.xe,&Seg.ye,&Status)!=6)
	      printf("Error in D: %s\n",msg);
	    Seg.Flag=Status;
	    AddDirectSegment(Id,&Seg);	    
	      break;
	  }
	case 'I':
	  {
	    int Status;
	    long Id;

	    if (sscanf(msg,"I %ld %ld %ld %ld %ld %d",
		       &Id,&Seg.xi,&Seg.yi,&Seg.xe,&Seg.ye,&Status)!=6)
	      printf("Error in I: %s\n",msg);
	    Seg.Flag=Status;
	    AddInverseSegment(Id,&Seg);	    
	      break;
	  }
	case 'L':
	  {
	    double ComputeTime;
	    int Status;
	    long Id;
	    int AddLevelPoint(int Id,struct Segment *Seg,double ComputeTime);

	    if (sscanf(msg,"L %ld %ld %ld %ld %ld %d %le %le %le %lf",
		       &Id,&Seg.xi,&Seg.yi,&Seg.xe,&Seg.ye,
		       &Status,&Seg.x,&Seg.y,&Seg.Sigma,&ComputeTime)!=10)
	      printf("Error in :%s\n",msg);
	    Seg.Flag=Status;
	    AddLevelPoint(Id,&Seg,ComputeTime);
	    break;
	  }
	case 'M':
	  RedrawMachines(msg+1);
	  break;
	case 'V':
	  {
	    double re,im;
	    if (sscanf(msg,"V %lf %lf",&re,&im)!=2)
	      printf("Error in :%s\n",msg);
	    else
	      AddEigenValue(re,im);
	    break;
	  }
	default:
	  printf("Received Message: %s\n",msg);
	}
      msg=strtok(NULL,"\n");
    }
  if (k)
    {
      memcpy(Buffer,&Buffer[SOCK_BUF_SIZE-k],k*sizeof(char));
    }
}




int SendMainMaster(char *s)
{
  if (SockNum==-1)
    printf("Write Socket: %s\n",s);
  else
    write(SockNum,s,strlen(s));
}

/* Communication Functions */
static char Buffer[200];


#define CREATE_LEVEL_CURVE    'C'
#define COMPUTE_S_OF_Z        'S'

int CreateLevelCurve(float Re,float Im,
		     float LevelCurve,
		     float MeshSize,
		     char Method,
		     float SigmaMinPrec,
		     long SigmaMinMaxIt,
		     float SplitPrec,
		     char Sym,
		     int Slices)
{
  int IMethod;

  switch(Method)
    {
    case 'L':
      IMethod=LANCZOS;
      break;
    case 'D':
      IMethod=DAVIDSON;
      break;
    case 'T':
      IMethod=TRACEMIN;
      break;
    case 'A':
      IMethod=ARNOLDI;
      break;
    }
  sprintf(Buffer,"C %le %le %le %le %le %ld %le %d %d %d\n",
	  Re,Im,MeshSize,
	  LevelCurve,SigmaMinPrec,SigmaMinMaxIt,
	  SplitPrec,(int) (Sym!=0),IMethod,Slices);
  SendMainMaster(Buffer);
}

int Compute_Sz(float Re,float Im,char Method,
	       float SigmaMinPrec,long SigmaMinMaxIt)
{
  int IMethod;
  switch(Method)
    {
    case 'L':
      IMethod=LANCZOS;
      break;
    case 'D':
      IMethod=DAVIDSON;
      break;
    case 'T':
      IMethod=TRACEMIN;
      break;
    case 'A':
      IMethod=ARNOLDI;
      break;
    }
  sprintf(Buffer,"%c %e %e %e %ld %d",(char) COMPUTE_S_OF_Z,
	  Re,Im,SigmaMinPrec,SigmaMinMaxIt,IMethod);
  SendMainMaster(Buffer);
}

void ComputeEigs(double Re,double Im,long nev,long ncv,
		 long maxit,double eps)
{
  sprintf(Buffer,"V %e %e %d %d %d %e",Re,Im,nev,ncv,maxit,eps);
  SendMainMaster(Buffer);
}



/***************************** Drawing Level Curves *********************************/

struct GLevelCurve
{
  int Id;
  char Flag;
  char Method;
  double Tau;
  double Sigma,SigmaPrec,SplitPrec;
  double CreatedTime,LastModified;
  long SigmaMaxIt;
  char OrbitFlags;
  void *Orbit; /* Liste des points */
  struct GLevelCurve *Next;
  int Priority;
  int MinP,MaxP;
  char DrawTriangles;
  int Limit2Triangles;
  GtkWidget *PriorityWidget;
  GtkWidget *MinPWidget,*MaxPWidget;
  GtkWidget *CurveColorDrawingArea;
  GtkWidget *InternalPointsColorDrawingArea;
  GtkWidget *ExternalPointsColorDrawingArea;
  GtkWidget *UncertainPointsColorDrawingArea;
  GtkWidget *SegmentColorDrawingArea;  
  GdkColor CurveColor;
  GdkColor InternalPointsColor;
  GdkColor ExternalPointsColor;
  GdkColor UncertainPointsColor;
  GdkColor SegmentColor;
};

static struct GLevelCurve *Curves=NULL;

void CreateGLevelCurve(long Id,char Flag,char Method,double Tau,
		       double Sigma,double SigmaPrec,double SplitPrec,
		       long SigmaMaxIt,double ti,GtkWidget *drawing_area)
{
  struct GLevelCurve *tmp;

  tmp=Curves;
  Curves=(struct GLevelCurve *)malloc(sizeof(struct GLevelCurve));
  Curves->Next=tmp;
  Curves->Id=Id;
  Curves->Flag=Flag;
  Curves->Method=Method;
  Curves->Tau=Tau;
  Curves->Sigma=Sigma;
  Curves->SigmaPrec=SigmaPrec;
  Curves->SplitPrec=SplitPrec;
  Curves->SigmaMaxIt=SigmaMaxIt;
  Curves->Orbit=CreateEmptyList();
  Curves->DrawTriangles=1;
  Curves->Limit2Triangles=0;
  Curves->Priority=0;
  Curves->MinP=1;
  Curves->MaxP=Machines;
  Curves->CreatedTime=Curves->LastModified=ti;
  srand(time(0));
  /* Default Curve Color */
  Curves->CurveColor.red=rand();
  Curves->CurveColor.blue=rand();
  Curves->CurveColor.green=rand();
  Curves->CurveColor.pixel=(gulong)(Curves->CurveColor.red*65536
				    +Curves->CurveColor.green*256
				    +Curves->CurveColor.blue);
  gdk_color_alloc(gtk_widget_get_colormap(drawing_area),
		  &Curves->CurveColor);
  /* Default internal Color */
  Curves->InternalPointsColor.red=rand();
  Curves->InternalPointsColor.blue=rand();
  Curves->InternalPointsColor.green=rand();
  Curves->InternalPointsColor.pixel=(gulong)(Curves->InternalPointsColor.red*65536
					     +Curves->InternalPointsColor.green*256
					     +Curves->InternalPointsColor.blue);
  gdk_color_alloc(gtk_widget_get_colormap(drawing_area),
		  &Curves->InternalPointsColor);
  /* Default external Color */
  Curves->ExternalPointsColor.red=rand();
  Curves->ExternalPointsColor.blue=rand();
  Curves->ExternalPointsColor.green=rand();
  Curves->ExternalPointsColor.pixel=(gulong)(Curves->ExternalPointsColor.red*65536
					     +Curves->ExternalPointsColor.green*256
					     +Curves->ExternalPointsColor.blue);
  gdk_color_alloc(gtk_widget_get_colormap(drawing_area),
		  &Curves->ExternalPointsColor);
  /* Default segment Color */
  Curves->SegmentColor.red=rand();
  Curves->SegmentColor.blue=rand();
  Curves->SegmentColor.green=rand();
  Curves->SegmentColor.pixel=(gulong)(Curves->SegmentColor.red*65536
				      +Curves->SegmentColor.green*256
				      +Curves->SegmentColor.blue);
  gdk_color_alloc(gtk_widget_get_colormap(drawing_area),
		  &Curves->SegmentColor);
  /* Default uncertain Color  */
  Curves->UncertainPointsColor.red=65535;
  Curves->UncertainPointsColor.blue=0;
  Curves->UncertainPointsColor.green=0;
  Curves->UncertainPointsColor.pixel=(gulong)(Curves->UncertainPointsColor.red*65536
					      +Curves->UncertainPointsColor.green*256
					      +Curves->UncertainPointsColor.blue);
  gdk_color_alloc(gtk_widget_get_colormap(drawing_area),
		   &Curves->UncertainPointsColor);
}

int PrintSeg(void *Seg,void *file)
{
  struct Segment *S=(struct Segment *)Seg;
  FILE *fp=(FILE *)file;
  if (S->Flag & Zi_CONV)
    fprintf(fp,"Zi_CONV ");
  else
    fprintf(fp,"******* ");
  if (S->Flag & Ze_CONV)
    fprintf(fp,"Ze_CONV ");
  else
    fprintf(fp,"******* ");
  if (S->Flag & Zl_INCL)
    fprintf(fp,"Zl_INCL ");
  else
    fprintf(fp,"******* ");
  if (S->Flag & Zl_CONV)
    fprintf(fp,"Zl_CONV ");
  else
    fprintf(fp,"******* ");

  fprintf(fp,"%ld %ld %ld %ld %le %le %le\n",
	  S->xi,S->yi,S->xe,S->ye,S->x,S->y,S->Sigma);
  return 0;
}

int SaveLevelCurves(char *filename)
{
  FILE *fp;
  struct GLevelCurve *tmp=Curves;

  fp=fopen(filename,"wt+");
  if (fp==NULL)
    return 0;
  while (tmp)
    {
      fprintf(fp,"Level Curve: %e\n",tmp->Sigma);
      fprintf(fp,"Mesh Size: %e\n",tmp->Tau);
      switch(tmp->Method)
	{
	case LANCZOS:
	  fprintf(fp,"Numerical Algorithm:Lanczos\n");
	  break;
	case DAVIDSON:
	  fprintf(fp,"Numerical Algorithm:Davidson\n");
	  break;
	case TRACEMIN:
	  fprintf(fp,"Numerical Algorithm:Trace Minimization\n");
	  break;
	case ARNOLDI:
	  fprintf(fp,"Numerical Algorithm:Arnoldi\n");
	  break;
	}
      fprintf(fp,"Numerical precision: %e\n",tmp->SigmaPrec);
      fprintf(fp,"Maximum Iteration  : %ld\n",tmp->SigmaMaxIt);
      fprintf(fp,"Bisection precision: %e\n",tmp->SplitPrec);
      fprintf(fp,"Created time: %.0f\n",tmp->CreatedTime);
      fprintf(fp,"Last Modified: %.0f\n",tmp->LastModified);
      ForAllData(tmp->Orbit,PrintSeg,(void *)fp);
      fprintf(fp,"----------\n\n\n\n");
      tmp=tmp->Next;
    }
  fclose(fp);
  return 1;
}

int LoadLevelCurves(char *filename)
{
  FILE *fp,*fp1;
  char s[100];
  double Sigma;
  long it;
  struct GLevelCurve *tmp;
  struct Segment Seg;

  fp=fopen(filename,"rt");
  if (fp==NULL)
    return;
  while(fgets(s,100,fp))
    {
      if (s[0]=='-' || s[0]=='\0' || s[0]==' ')
	continue;
      if (sscanf(s,"Level Curve: %le\n",&Sigma)==1)
	{
	  tmp=Curves;
	  Curves=(struct GLevelCurve *)malloc(sizeof(struct GLevelCurve));
	  Curves->Next=tmp;
	  Curves->Id=-1;
	  Curves->Priority=0;
	  Curves->MinP=Curves->MaxP=-1;
	  Curves->Orbit=CreateEmptyList();
	  Curves->Sigma=Sigma;
	  /* Default Curve Color */
	  Curves->CurveColor.red=rand();
	  Curves->CurveColor.blue=rand();
	  Curves->CurveColor.green=rand();
	  Curves->CurveColor.pixel=(gulong)(Curves->CurveColor.red*65536
					    +Curves->CurveColor.green*256
					    +Curves->CurveColor.blue);
	  gdk_color_alloc(gtk_widget_get_colormap( (GtkWidget *) LevelDrawingArea),
			  &Curves->CurveColor);
	  /* Default internal Color */
	  Curves->InternalPointsColor.red=rand();
	  Curves->InternalPointsColor.blue=rand();
	  Curves->InternalPointsColor.green=rand();
	  Curves->InternalPointsColor.pixel=(gulong)(Curves->InternalPointsColor.red*65536
						     +Curves->InternalPointsColor.green*256
						     +Curves->InternalPointsColor.blue);
	  gdk_color_alloc(gtk_widget_get_colormap((GtkWidget *) LevelDrawingArea),
			  &Curves->InternalPointsColor);
	  /* Default external Color */
	  Curves->ExternalPointsColor.red=rand();
	  Curves->ExternalPointsColor.blue=rand();
	  Curves->ExternalPointsColor.green=rand();
	  Curves->ExternalPointsColor.pixel=(gulong)(Curves->ExternalPointsColor.red*65536
						     +Curves->ExternalPointsColor.green*256
						     +Curves->ExternalPointsColor.blue);
	  gdk_color_alloc(gtk_widget_get_colormap((GtkWidget *) LevelDrawingArea),
			  &Curves->ExternalPointsColor);
	  /* Default segment Color */
	  Curves->SegmentColor.red=rand();
	  Curves->SegmentColor.blue=rand();
	  Curves->SegmentColor.green=rand();
	  Curves->SegmentColor.pixel=(gulong)(Curves->SegmentColor.red*65536
					      +Curves->SegmentColor.green*256
					      +Curves->SegmentColor.blue);
	  gdk_color_alloc(gtk_widget_get_colormap((GtkWidget *) LevelDrawingArea),
			  &Curves->SegmentColor);
	  /* Default uncertain Color  */
	  Curves->UncertainPointsColor.red=65535;
	  Curves->UncertainPointsColor.blue=0;
	  Curves->UncertainPointsColor.green=0;
	  Curves->UncertainPointsColor.pixel=(gulong)(Curves->UncertainPointsColor.red*65536
						      +Curves->UncertainPointsColor.green*256
						      +Curves->UncertainPointsColor.blue);
	  gdk_color_alloc(gtk_widget_get_colormap((GtkWidget *) LevelDrawingArea),
			  &Curves->UncertainPointsColor);

	  continue;
	}
      if (sscanf(s,"Mesh Size: %le\n",&Sigma)==1)
	{
	  Curves->Tau=Sigma;
	  continue;
	}
      if (sscanf(s,"Numerical precision: %le\n",&Sigma)==1)
	{
	  Curves->SigmaPrec=Sigma;
	  continue;
	}
      if (sscanf(s,"Maximum Iteration  : %ld\n",&it)==1)
	{
	  Curves->SigmaMaxIt=it;
	  continue;
	}
      if (sscanf(s,"Bisection precision: %le\n",&Sigma)==1)
	{
	  Curves->SplitPrec=Sigma;
	  continue;
	}
      if (sscanf(s,"Created time: %f\n",&Sigma)==1)
	{
	  Curves->CreatedTime=Sigma;
	  continue;
	}
      if (sscanf(s,"Last Modified: %f\n",&Sigma)==1)
	{
	  Curves->LastModified=Sigma;
	  continue;
	}
      if ((s[0]=='Z' || s[0]=='*') &&  
	  sscanf(&s[32],"%ld %ld %ld %ld %le %le %le\n",
		 &Seg.xi,&Seg.yi,&Seg.xe,&Seg.ye,
		 &Seg.x,&Seg.y,&Seg.Sigma)==7)
	{
	  Seg.Flag=0;
	  if (s[0]=='Z')
	    Seg.Flag |=Zi_CONV;
	  if (s[8]=='Z')
	    Seg.Flag |= Ze_CONV;
	  if (s[16]=='Z')
	    Seg.Flag |=Zl_INCL;
	  if (s[24]=='Z')
	    Seg.Flag |= Zl_CONV;
	  AddFirst(&Seg,sizeof(struct Segment),Curves->Orbit);
	  continue;
	}
      switch(s[20])
	{
	case 'L':
	  Curves->Method=LANCZOS;break;
	case 'D':
	  Curves->Method=DAVIDSON;break;
	case 'T':
	  Curves->Method=TRACEMIN;break;
	case 'A':
	  Curves->Method=ARNOLDI;break;
	}
    }
  fclose(fp);
  RedrawAllCurves();
}


int AddDirectSegment(int Id,struct Segment *s)
{
  struct GLevelCurve *tmp;

  tmp=Curves;
  while (tmp)
    {
      if (tmp->Id==Id)
	break;
      tmp=tmp->Next;
    }
  if (tmp)
    {
      if (tmp->DrawTriangles)
	{
	  GdkRectangle update_rect;
	  double TauCosPI_3,TauSinPI_3;
	  double xi,yi,xe,ye;
	  gint x00,x01,y00,y01,x10,x11,y10,y11;
	  struct Segment *S1;
	  
	  if (LevelGC==NULL)
	    LevelGC=gdk_gc_new(LevelPixmap);
	  TauCosPI_3=tmp->Tau*COSPI_3;
	  TauSinPI_3=tmp->Tau*SINPI_3;
	  xi=s->xi*tmp->Tau+s->yi*TauCosPI_3;
	  yi=s->yi*TauSinPI_3;
	  xe=s->xe*tmp->Tau+s->ye*TauCosPI_3;
	  ye=s->ye*TauSinPI_3;
	  x10=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
		      /(HRuler->upper-HRuler->lower)*(xi-HRuler->lower));
	  y10=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
		      *(yi-VRuler->lower)/(VRuler->upper-VRuler->lower));
	  x11=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
		      /(HRuler->upper-HRuler->lower)*(xe-HRuler->lower));
	  y11=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
		      *(ye-VRuler->lower)/(VRuler->upper-VRuler->lower));
	  gdk_gc_set_foreground(LevelGC,&tmp->SegmentColor);
	  gdk_draw_line(LevelPixmap,LevelGC,x10,y10,x11,y11);
	  S1=GetFirstDataAdress(tmp->Orbit);
	  if (S1)
	    {
	      xi=S1->xi*tmp->Tau+S1->yi*TauCosPI_3;
	      yi=S1->yi*TauSinPI_3;
	      xe=S1->xe*tmp->Tau+S1->ye*TauCosPI_3;
	      ye=S1->ye*TauSinPI_3;
	      x00=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
			  /(HRuler->upper-HRuler->lower)*(xi-HRuler->lower));
	      y00=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
			  *(yi-VRuler->lower)/(VRuler->upper-VRuler->lower));
	      x01=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
			  /(HRuler->upper-HRuler->lower)*(xe-HRuler->lower));
	      y01=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
			  *(ye-VRuler->lower)/(VRuler->upper-VRuler->lower));
	      gdk_draw_line(LevelPixmap,LevelGC,x00,y00,x10,y10);
	      gdk_draw_line(LevelPixmap,LevelGC,x01,y01,x11,y11);
	    }
	  if (s->Flag & Zi_CONV)
	    gdk_gc_set_foreground(LevelGC,&tmp->InternalPointsColor);
	  else
	    gdk_gc_set_foreground(LevelGC,&tmp->UncertainPointsColor);
	  gdk_draw_line(LevelPixmap,LevelGC,x10-2,y10-2,x10+2,y10+2);
	  gdk_draw_line(LevelPixmap,LevelGC,x10-2,y10+2,x10+2,y10-2);
	  if (s->Flag & Ze_CONV)
	    gdk_gc_set_foreground(LevelGC,&tmp->ExternalPointsColor);
	  else
	    gdk_gc_set_foreground(LevelGC,&tmp->UncertainPointsColor);   
	  gdk_draw_line(LevelPixmap,LevelGC,x11-2,y11-2,x11+2,y11+2);
	  gdk_draw_line(LevelPixmap,LevelGC,x11-2,y11+2,x11+2,y11-2);
	  update_rect.x=0;
	  update_rect.y=0;
	  update_rect.width=GTK_WIDGET(LevelDrawingArea)->allocation.width;
	  update_rect.height=GTK_WIDGET(LevelDrawingArea)->allocation.height;
	  gtk_widget_draw(GTK_WIDGET(LevelDrawingArea), &update_rect);
	}
      AddFirst(s,sizeof(struct Segment),tmp->Orbit);
    }
}

int AddInverseSegment(int Id,struct Segment *s)
{
  struct GLevelCurve *tmp;

  tmp=Curves;
  while (tmp)
    {
      if (tmp->Id==Id)
	break;
      tmp=tmp->Next;
    }
  if (tmp)
    {
      if (tmp->DrawTriangles)
	{
	  GdkRectangle update_rect;
	  double TauCosPI_3,TauSinPI_3;
	  double xi,yi,xe,ye;
	  gint x00,x01,y00,y01,x10,x11,y10,y11;
	  struct Segment *S1;
	  
	  if (LevelGC==NULL)
	    LevelGC=gdk_gc_new(LevelPixmap);
	  TauCosPI_3=tmp->Tau*COSPI_3;
	  TauSinPI_3=tmp->Tau*SINPI_3;
	  xi=s->xi*tmp->Tau+s->yi*TauCosPI_3;
	  yi=s->yi*TauSinPI_3;
	  xe=s->xe*tmp->Tau+s->ye*TauCosPI_3;
	  ye=s->ye*TauSinPI_3;
	  x10=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
		      /(HRuler->upper-HRuler->lower)*(xi-HRuler->lower));
	  y10=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
		      *(yi-VRuler->lower)/(VRuler->upper-VRuler->lower));
	  x11=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
		      /(HRuler->upper-HRuler->lower)*(xe-HRuler->lower));
	  y11=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
		      *(ye-VRuler->lower)/(VRuler->upper-VRuler->lower));
	  gdk_gc_set_foreground(LevelGC,&tmp->SegmentColor);
	  gdk_draw_line(LevelPixmap,LevelGC,x10,y10,x11,y11);
	  S1=GetLastDataAdress(tmp->Orbit);
	  if (S1)
	    {
	      xi=S1->xi*tmp->Tau+S1->yi*TauCosPI_3;
	      yi=S1->yi*TauSinPI_3;
	      xe=S1->xe*tmp->Tau+S1->ye*TauCosPI_3;
	      ye=S1->ye*TauSinPI_3;
	      x00=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
			  /(HRuler->upper-HRuler->lower)*(xi-HRuler->lower));
	      y00=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
			  *(yi-VRuler->lower)/(VRuler->upper-VRuler->lower));
	      x01=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
			  /(HRuler->upper-HRuler->lower)*(xe-HRuler->lower));
	      y01=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
			  *(ye-VRuler->lower)/(VRuler->upper-VRuler->lower));
	      gdk_draw_line(LevelPixmap,LevelGC,x00,y00,x10,y10);
	      gdk_draw_line(LevelPixmap,LevelGC,x01,y01,x11,y11);
	    }
	  if (s->Flag & Zi_CONV)
	    gdk_gc_set_foreground(LevelGC,&tmp->InternalPointsColor);
	  else
	    gdk_gc_set_foreground(LevelGC,&tmp->UncertainPointsColor);
	  gdk_draw_line(LevelPixmap,LevelGC,x10-2,y10-2,x10+2,y10+2);
	  gdk_draw_line(LevelPixmap,LevelGC,x10-2,y10+2,x10+2,y10-2);
	  if (s->Flag & Ze_CONV)
	    gdk_gc_set_foreground(LevelGC,&tmp->ExternalPointsColor);
	  else
	    gdk_gc_set_foreground(LevelGC,&tmp->UncertainPointsColor);   
	  gdk_draw_line(LevelPixmap,LevelGC,x11-2,y11-2,x11+2,y11+2);
	  gdk_draw_line(LevelPixmap,LevelGC,x11-2,y11+2,x11+2,y11-2);
	  update_rect.x=0;
	  update_rect.y=0;
	  update_rect.width=GTK_WIDGET(LevelDrawingArea)->allocation.width;
	  update_rect.height=GTK_WIDGET(LevelDrawingArea)->allocation.height;
	  gtk_widget_draw(GTK_WIDGET(LevelDrawingArea), &update_rect);
	}
      AddLast(s,sizeof(struct Segment),tmp->Orbit);
    }
}

int UpdateData(void *data,void *Seg)
{
  struct Segment *Sdata=(struct Segment *)data;
  struct Segment *SSeg=(struct Segment *)Seg;

  if (Sdata->Flag & Zl_INCL) /* If segment has level point enclosed */
    return 0;
  if (Sdata->xi==SSeg->xi && Sdata->yi==SSeg->yi &&
      Sdata->xe==SSeg->xe && Sdata->ye==SSeg->ye)
    {
      Sdata->Flag=SSeg->Flag;
      Sdata->x=SSeg->x;
      Sdata->y=SSeg->y;
      Sdata->Sigma=SSeg->Sigma;
      return 1;
    }
  return 0;
}

int PrintSegment(void *data,void *pointer)
{
  struct Segment *s=(struct Segment *)data;

  printf("%d %d %d %d %d %e %e %e\n",
	 s->xi,s->yi,s->xe,s->ye,s->Flag,s->x,s->y,s->Sigma);
  return 0;
}

int AddLevelPoint(int Id,struct Segment *Seg,double ComputeTime)
{
  struct GLevelCurve *tmp;
  
  tmp=Curves;
  while (tmp)
    {
      if (tmp->Id==Id)
	break;
      tmp=tmp->Next;
    }
  if (tmp)
    {
      tmp->LastModified=ComputeTime;
      ForAllData(tmp->Orbit,UpdateData,Seg);
      DrawLevelCurve(tmp);
    }
}


static int DrawTriangle(void *data,void *pointer)
{  
  struct Segment *Seg=(struct Segment *)data;
  struct GLevelCurve *Lev=(struct GLevelCurve *)pointer;
  double TauCosPI_3;
  double TauSinPI_3;
  gint x0,y0,x1,y1;
  static gint x00,y00,x01,y01;
  static char c;
  double xi,yi,xe,ye;

  if (data==NULL && pointer ==NULL)
    {
      c=0;
      return;
    }
  TauCosPI_3=Lev->Tau*COSPI_3;
  TauSinPI_3=Lev->Tau*SINPI_3;
  xi=Seg->xi*Lev->Tau+Seg->yi*TauCosPI_3;
  yi=Seg->yi*TauSinPI_3;
  xe=Seg->xe*Lev->Tau+Seg->ye*TauCosPI_3;
  ye=Seg->ye*TauSinPI_3;
  x0=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
	     /(HRuler->upper-HRuler->lower)*(xi-HRuler->lower));
  y0=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
	     *(yi-VRuler->lower)/(VRuler->upper-VRuler->lower));
  x1=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
	     /(HRuler->upper-HRuler->lower)*(xe-HRuler->lower));
  y1=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
	     *(ye-VRuler->lower)/(VRuler->upper-VRuler->lower));
  gdk_gc_set_foreground(LevelGC,&Lev->SegmentColor);
  gdk_draw_line(LevelPixmap,LevelGC,x0,y0,x1,y1);
  if (c)
    {
      gdk_draw_line(LevelPixmap,LevelGC,x0,y0,x00,y00);
      gdk_draw_line(LevelPixmap,LevelGC,x01,y01,x1,y1);
    }
  c=1;
  x00=x0;
  y00=y0;
  x01=x1;
  y01=y1;
  if (Seg->Flag & Zi_CONV)
    gdk_gc_set_foreground(LevelGC,&Lev->InternalPointsColor);
  else
    gdk_gc_set_foreground(LevelGC,&Lev->UncertainPointsColor);
  gdk_draw_line(LevelPixmap,LevelGC,x0-2,y0-2,x0+2,y0+2);
  gdk_draw_line(LevelPixmap,LevelGC,x0-2,y0+2,x0+2,y0-2);
  if (Seg->Flag & Ze_CONV)
    gdk_gc_set_foreground(LevelGC,&Lev->ExternalPointsColor);
  else
     gdk_gc_set_foreground(LevelGC,&Lev->UncertainPointsColor);   
  gdk_draw_line(LevelPixmap,LevelGC,x1-2,y1-2,x1+2,y1+2);
  gdk_draw_line(LevelPixmap,LevelGC,x1-2,y1+2,x1+2,y1-2);
  return 0;
}


DrawTriangles(struct GLevelCurve *Lev)
{
  GdkRectangle update_rect;
  if (Lev->DrawTriangles==0)
    return;
  if (LevelGC==NULL)
    LevelGC=gdk_gc_new(LevelPixmap);
  DrawTriangle(NULL,NULL); /* Used to initialise local data structures */
  ForAllData(Lev->Orbit,DrawTriangle,Lev);
  update_rect.x=0;
  update_rect.y=0;
  update_rect.width=GTK_WIDGET(LevelDrawingArea)->allocation.width;
  update_rect.height=GTK_WIDGET(LevelDrawingArea)->allocation.height;
  gtk_widget_draw(GTK_WIDGET(LevelDrawingArea), &update_rect);
}


static int DrawSegment(void *data,void *pointer)
{  
  struct Segment *Seg=(struct Segment *)data;
  struct GLevelCurve *Lev=(struct GLevelCurve *)pointer;
  gint x0,y0;
  static gint x00,y00;
  static char c,Status;

  if (data==NULL && pointer ==NULL)
    {
      c=0;
      return;
    }
  x0=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
	     /(HRuler->upper-HRuler->lower)*(Seg->x-HRuler->lower));
  y0=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
	     *(Seg->y-VRuler->lower)/(VRuler->upper-VRuler->lower));
  if (c && (Status & Zl_INCL) && (Seg->Flag & Zl_INCL))
    {
      if (Seg->Flag & Zl_CONV)
	gdk_gc_set_foreground(LevelGC,&Lev->CurveColor);
      else
	gdk_gc_set_foreground(LevelGC,&Lev->UncertainPointsColor);   
      gdk_draw_line(LevelPixmap,LevelGC,x0,y0,x00,y00);
    }
  Status=Seg->Flag;
  c=1;
  x00=x0;
  y00=y0;
  return 0;
}


DrawLevelCurve(struct GLevelCurve *Lev)
{
  GdkRectangle update_rect;
  if (LevelGC==NULL)
    LevelGC=gdk_gc_new(LevelPixmap);
  DrawSegment(NULL,NULL); /* Used to initialise local data structures */
  ForAllData(Lev->Orbit,DrawSegment,Lev);
  update_rect.x=0;
  update_rect.y=0;
  update_rect.width=GTK_WIDGET(LevelDrawingArea)->allocation.width;
  update_rect.height=GTK_WIDGET(LevelDrawingArea)->allocation.height;
  gtk_widget_draw(GTK_WIDGET(LevelDrawingArea), &update_rect);
}

RedrawAllCurves()
{
  struct GLevelCurve *tmp=Curves;
  if (LevelDrawingArea==NULL)
    return;
  gdk_draw_rectangle (LevelPixmap,
                      GTK_WIDGET(LevelDrawingArea)->style->black_gc,
                      TRUE,
                      0, 0,
                      GTK_WIDGET(LevelDrawingArea)->allocation.width,
                      GTK_WIDGET(LevelDrawingArea)->allocation.width);
  while (tmp)
    {
      DrawTriangles(tmp);
      DrawLevelCurve(tmp);
      tmp=tmp->Next;
    }
  DrawEigenValues();
}

RedrawMachines(char *m)
{
  static char mach[100]="";
  static GdkColor Green,Red;
  char *p;
  struct GLevelCurve *Tmp;
  GdkRectangle update_rect;
  long x=2;
  long y=2;

  if (m==NULL && mach[0]=='\0')
    return;

  if (MachinesGC==NULL)
    MachinesGC=gdk_gc_new(MachinePixmap);
 
  update_rect.y = y;
  update_rect.width = 6;
  update_rect.height = 6;

  if (mach[0]=='\0')
    {
      /* Create Green and Red Colors */
      Green.red=0;
      Green.blue=0;
      Green.green=65535;
      Green.pixel=(gulong)(Green.red*65536
			   +Green.green*256
			   +Green.blue);
      gdk_color_alloc(gtk_widget_get_colormap(GTK_WIDGET(MachineDrawingArea)),
		      &Green);
      Red.red=65535;
      Red.blue=0;
      Red.green=0;
      Red.pixel=(gulong)(Red.red*65536
			 +Red.green*256
			 +Red.blue);
      gdk_color_alloc(gtk_widget_get_colormap(GTK_WIDGET(MachineDrawingArea)),
		      &Red);
    }
  
  gdk_draw_rectangle (MachinePixmap,
                      GTK_WIDGET(MachineDrawingArea)->style->black_gc,
                      TRUE,
                      0, 0,
                      GTK_WIDGET(MachineDrawingArea)->allocation.width,
                      GTK_WIDGET(MachineDrawingArea)->allocation.width);
  if (m)
    strcpy(mach,m);
  p=mach;
  Machines=0;
  while (*p!='\0')
    {
      update_rect.x = 10*x;
      if (*p=='-')
	{
	  Machines++;
	  gdk_gc_set_foreground(MachinesGC,&Green);
	  gdk_draw_rectangle (MachinePixmap,
			      MachinesGC,
			      TRUE,
			      update_rect.x, update_rect.y,
			      update_rect.width, update_rect.height);
	}
      else
	if (*p=='u')
	  {
	    Machines++;
	    gdk_gc_set_foreground(MachinesGC,&Red);
	    gdk_draw_rectangle (MachinePixmap,
				MachinesGC,
				TRUE,
				update_rect.x, update_rect.y,
				update_rect.width, update_rect.height);
	  }
	else
	  {
	    int Id=*p-'A'+1;
	    Machines++;
	    Tmp=Curves;
	    while(Tmp)
	      {
		if (Tmp->Id==Id)
		  break;
		Tmp=Tmp->Next;
	      }
	    if (Tmp!=NULL)
	      gdk_gc_set_foreground(MachinesGC,&Tmp->CurveColor);
	    else
	      gdk_gc_set_foreground(MachinesGC,&Red);
	    gdk_draw_rectangle (MachinePixmap,
				MachinesGC,
				TRUE,
				update_rect.x, update_rect.y,
				update_rect.width, update_rect.height); 
	  }
      gtk_widget_draw (GTK_WIDGET(MachineDrawingArea), &update_rect);
      x++;
      p++;
    }
}




/*****************************************Level Curve Properties*********************************/



static void Triang(GtkWidget *widget,struct GLevelCurve *Data)
{
  Data->DrawTriangles=GTK_TOGGLE_BUTTON(widget)->active;
  RedrawAllCurves();
}
static void Limit2Triang(GtkWidget *widget,struct GLevelCurve *Data)
{
  char s[20];
  Data->Limit2Triangles=!GTK_TOGGLE_BUTTON(widget)->active;
  sprintf(s,"T %d %d\n",Data->Id,Data->Limit2Triangles);
  SendMainMaster(s);
}

static void MethodName(int Method,char *s,int n)
{
  switch (Method)
    {
    case LANCZOS:
      strncpy(s,"Lanczos",n);
      break;
    case ARNOLDI:
      strncpy(s,"Arnoldi (ARPACK)",n);
      break;
    case TRACEMIN:
      strncpy(s,"Trace Minimization",n);
      break;
    case DAVIDSON:
      strncpy(s,"Davidson",n);
      break;
    }
}

void PriorityChanged(GtkWidget *widget,struct GLevelCurve *Lev)
{
  int p;
  char s[20];
  if (sscanf(gtk_entry_get_text(GTK_ENTRY(widget)),"%d",&p)!=1)
    return;
  Lev->Priority=p;
  sprintf(s,"N %d %d\n",(int)Lev->Id,(int)Lev->Priority);
  SendMainMaster(s);
}

void MinPChanged(GtkWidget *widget,struct GLevelCurve *Lev)
{
  int p;
  char s[20];
  if (sscanf(gtk_entry_get_text(GTK_ENTRY(widget)),"%d",&p)!=1)
    return;
  Lev->MinP=p;
  sprintf(s,"E %d %d %d\n",(int)Lev->Id,(int)Lev->MinP,(int)Lev->MaxP);
  SendMainMaster(s);
}
void MaxPChanged(GtkWidget *widget,struct GLevelCurve *Lev)
{
  int p;
  char s[20];
  if (sscanf(gtk_entry_get_text(GTK_ENTRY(widget)),"%d",&p)!=1)
    return;
  Lev->MaxP=p;
  sprintf(s,"E %d %d %d\n",(int)Lev->Id,(int)Lev->MinP,(int)Lev->MaxP);
  SendMainMaster(s);
}

#define C_COLOR      1   /* Curve Color */
#define S_COLOR      2   /* Segment Color */
#define I_COLOR      3   /* Internal Color */
#define E_COLOR      4   /* External Color */
#define U_COLOR      5   /* Uncertain Color */

static struct UserData
{
  struct GLevelCurve *Tmp;
  char which;
}In,Ex,Un,Cu,Se;


static void CurveColorChanged( GtkWidget         *widget,
			       struct GLevelCurve *Tmp)
{
  gdouble color[3];
  GdkColor gdk_color;
  GdkColormap *colormap;
  GtkColorSelection *colorsel=(GtkColorSelection *)widget;

  /* Get drawingarea colormap */


  colormap = gdk_window_get_colormap (Tmp->CurveColorDrawingArea->window);

  /* Get current color */
  
  gtk_color_selection_get_color (colorsel,color);

  /* Fit to a unsigned 16 bit integer (0..65535) and
   * insert into the GdkColor structure */

  gdk_color.red = (guint16)(color[0]*65535.0);
  gdk_color.green = (guint16)(color[1]*65535.0);
  gdk_color.blue = (guint16)(color[2]*65535.0);

  /* Allocate color */

  gdk_color_alloc (colormap, &gdk_color);

  /* Set window background color */
  gdk_window_set_background (Tmp->CurveColorDrawingArea->window, &gdk_color);
  gdk_window_clear (Tmp->CurveColorDrawingArea->window);
  Tmp->CurveColor=gdk_color;

  DrawLevelCurve(Tmp);
  RedrawMachines(NULL);
}

static void SegmentColorChanged( GtkWidget         *widget,
			       struct GLevelCurve *Tmp)
{
  gdouble color[3];
  GdkColor gdk_color;
  GdkColormap *colormap;
  GtkColorSelection *colorsel=(GtkColorSelection *)widget;

  /* Get drawingarea colormap */


  colormap = gdk_window_get_colormap (Tmp->SegmentColorDrawingArea->window);

  /* Get current color */
  
  gtk_color_selection_get_color (colorsel,color);

  /* Fit to a unsigned 16 bit integer (0..65535) and
   * insert into the GdkColor structure */

  gdk_color.red = (guint16)(color[0]*65535.0);
  gdk_color.green = (guint16)(color[1]*65535.0);
  gdk_color.blue = (guint16)(color[2]*65535.0);

  /* Allocate color */

  gdk_color_alloc (colormap, &gdk_color);

  /* Set window background color */
  gdk_window_set_background (Tmp->SegmentColorDrawingArea->window, &gdk_color);
  gdk_window_clear (Tmp->SegmentColorDrawingArea->window);
  Tmp->SegmentColor=gdk_color;

  DrawTriangles(Tmp);
}

static void InternalPointsColorChanged( GtkWidget         *widget,
			       struct GLevelCurve *Tmp)
{
  gdouble color[3];
  GdkColor gdk_color;
  GdkColormap *colormap;
  GtkColorSelection *colorsel=(GtkColorSelection *)widget;

  /* Get drawingarea colormap */


  colormap = gdk_window_get_colormap (Tmp->InternalPointsColorDrawingArea->window);

  /* Get current color */
  
  gtk_color_selection_get_color (colorsel,color);

  /* Fit to a unsigned 16 bit integer (0..65535) and
   * insert into the GdkColor structure */

  gdk_color.red = (guint16)(color[0]*65535.0);
  gdk_color.green = (guint16)(color[1]*65535.0);
  gdk_color.blue = (guint16)(color[2]*65535.0);

  /* Allocate color */

  gdk_color_alloc (colormap, &gdk_color);

  /* Set window background color */
  gdk_window_set_background (Tmp->InternalPointsColorDrawingArea->window, &gdk_color);
  gdk_window_clear (Tmp->InternalPointsColorDrawingArea->window);
  Tmp->InternalPointsColor=gdk_color;

  DrawTriangles(Tmp);
}

static void ExternalPointsColorChanged( GtkWidget         *widget,
			       struct GLevelCurve *Tmp)
{
  gdouble color[3];
  GdkColor gdk_color;
  GdkColormap *colormap;
  GtkColorSelection *colorsel=(GtkColorSelection *)widget;

  /* Get drawingarea colormap */


  colormap = gdk_window_get_colormap (Tmp->ExternalPointsColorDrawingArea->window);

  /* Get current color */
  
  gtk_color_selection_get_color (colorsel,color);

  /* Fit to a unsigned 16 bit integer (0..65535) and
   * insert into the GdkColor structure */

  gdk_color.red = (guint16)(color[0]*65535.0);
  gdk_color.green = (guint16)(color[1]*65535.0);
  gdk_color.blue = (guint16)(color[2]*65535.0);

  /* Allocate color */

  gdk_color_alloc (colormap, &gdk_color);

  /* Set window background color */
  gdk_window_set_background (Tmp->ExternalPointsColorDrawingArea->window, &gdk_color);
  gdk_window_clear (Tmp->ExternalPointsColorDrawingArea->window);
  Tmp->ExternalPointsColor=gdk_color;

  DrawTriangles(Tmp);
}

static void UncertainPointsColorChanged( GtkWidget         *widget,
			       struct GLevelCurve *Tmp)
{
  gdouble color[3];
  GdkColor gdk_color;
  GdkColormap *colormap;
  GtkColorSelection *colorsel=(GtkColorSelection *)widget;

  /* Get drawingarea colormap */


  colormap = gdk_window_get_colormap (Tmp->UncertainPointsColorDrawingArea->window);

  /* Get current color */
  
  gtk_color_selection_get_color (colorsel,color);

  /* Fit to a unsigned 16 bit integer (0..65535) and
   * insert into the GdkColor structure */

  gdk_color.red = (guint16)(color[0]*65535.0);
  gdk_color.green = (guint16)(color[1]*65535.0);
  gdk_color.blue = (guint16)(color[2]*65535.0);

  /* Allocate color */

  gdk_color_alloc (colormap, &gdk_color);

  /* Set window background color */
  gdk_window_set_background (Tmp->UncertainPointsColorDrawingArea->window, &gdk_color);
  gdk_window_clear (Tmp->UncertainPointsColorDrawingArea->window);
  Tmp->UncertainPointsColor=gdk_color;

  DrawTriangles(Tmp);
  DrawLevelCurve(Tmp);
}

static void CloseWin (GtkButton       *button,
		      gpointer         user_data)
{
  gtk_widget_destroy((GtkWidget *)user_data);
}


static gint CurveColorAreaEvent(GtkWidget *widget,
		GdkEvent  *event,
		struct GLevelCurve *Tmp)
{
  GtkWidget *colorseldlg;
  GtkWidget *colorsel;
  GtkWidget *ok_button;
  GtkWidget *cancel_button;
  GtkWidget *help_button;

  colorseldlg = gtk_color_selection_dialog_new("Changing Level Curve Color");
  colorsel = GTK_COLOR_SELECTION_DIALOG(colorseldlg)->colorsel;  

  ok_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->ok_button;
  cancel_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->cancel_button;
  help_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->help_button;
  gtk_widget_hide(cancel_button);
  gtk_widget_hide(help_button);
  gtk_signal_connect (GTK_OBJECT (ok_button), "clicked",
                      GTK_SIGNAL_FUNC (CloseWin),
                      colorseldlg);

  gtk_signal_connect(GTK_OBJECT(colorsel), "color_changed",
		     (GtkSignalFunc)CurveColorChanged, Tmp); 
  gtk_widget_show(colorseldlg);
  gdk_window_set_background (Tmp->CurveColorDrawingArea->window, 
			     &Tmp->CurveColor);
  gdk_window_clear (Tmp->CurveColorDrawingArea->window);    
}

static gint SegmentColorAreaEvent(GtkWidget *widget,
		GdkEvent  *event,
		struct GLevelCurve *Tmp)
{
  GtkWidget *colorseldlg;
  GtkWidget *colorsel;
  GtkWidget *ok_button;
  GtkWidget *cancel_button;
  GtkWidget *help_button;

  colorseldlg = gtk_color_selection_dialog_new("Changing Segment Color");
  colorsel = GTK_COLOR_SELECTION_DIALOG(colorseldlg)->colorsel;  
  
  ok_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->ok_button;
  cancel_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->cancel_button;
  help_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->help_button;
  gtk_widget_hide(cancel_button);
  gtk_widget_hide(help_button);
  gtk_signal_connect (GTK_OBJECT (ok_button), "clicked",
                      GTK_SIGNAL_FUNC (CloseWin),
                      colorseldlg);

  gtk_signal_connect(GTK_OBJECT(colorsel), "color_changed",
		     (GtkSignalFunc)SegmentColorChanged, Tmp); 
  gtk_widget_show(colorseldlg);
  gdk_window_set_background (Tmp->SegmentColorDrawingArea->window, 
			     &Tmp->SegmentColor);
  gdk_window_clear (Tmp->SegmentColorDrawingArea->window);    
}

static gint InternalPointsColorAreaEvent(GtkWidget *widget,
		GdkEvent  *event,
		struct GLevelCurve *Tmp)
{
  GtkWidget *colorseldlg;
  GtkWidget *colorsel;
  GtkWidget *ok_button;
  GtkWidget *cancel_button;
  GtkWidget *help_button;

  colorseldlg = gtk_color_selection_dialog_new("Changing Internal Points Color");
  colorsel = GTK_COLOR_SELECTION_DIALOG(colorseldlg)->colorsel;  

  ok_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->ok_button;
  cancel_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->cancel_button;
  help_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->help_button;
  gtk_widget_hide(cancel_button);
  gtk_widget_hide(help_button);
  gtk_signal_connect (GTK_OBJECT (ok_button), "clicked",
                      GTK_SIGNAL_FUNC (CloseWin),
                      colorseldlg);


  gtk_signal_connect(GTK_OBJECT(colorsel), "color_changed",
		     (GtkSignalFunc)InternalPointsColorChanged, Tmp); 
  gtk_widget_show(colorseldlg);
  gdk_window_set_background (Tmp->InternalPointsColorDrawingArea->window, 
			     &Tmp->InternalPointsColor);
  gdk_window_clear (Tmp->InternalPointsColorDrawingArea->window);    
}

static gint ExternalPointsColorAreaEvent(GtkWidget *widget,
		GdkEvent  *event,
		struct GLevelCurve *Tmp)
{
  GtkWidget *colorseldlg;
  GtkWidget *colorsel;
  GtkWidget *ok_button;
  GtkWidget *cancel_button;
  GtkWidget *help_button;


  colorseldlg = gtk_color_selection_dialog_new("Changing External Points Color");
  colorsel = GTK_COLOR_SELECTION_DIALOG(colorseldlg)->colorsel;  

  ok_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->ok_button;
  cancel_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->cancel_button;
  help_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->help_button;
  gtk_widget_hide(cancel_button);
  gtk_widget_hide(help_button);
  gtk_signal_connect (GTK_OBJECT (ok_button), "clicked",
                      GTK_SIGNAL_FUNC (CloseWin),
                      colorseldlg);

  gtk_signal_connect(GTK_OBJECT(colorsel), "color_changed",
		     (GtkSignalFunc)ExternalPointsColorChanged, Tmp); 
  gtk_widget_show(colorseldlg);
  gdk_window_set_background (Tmp->ExternalPointsColorDrawingArea->window, 
			     &Tmp->ExternalPointsColor);
  gdk_window_clear (Tmp->ExternalPointsColorDrawingArea->window);    
}

static gint UncertainPointsColorAreaEvent(GtkWidget *widget,
		GdkEvent  *event,
		struct GLevelCurve *Tmp)
{
  GtkWidget *colorseldlg;
  GtkWidget *colorsel;
  GtkWidget *ok_button;
  GtkWidget *cancel_button;
  GtkWidget *help_button;


  colorseldlg = gtk_color_selection_dialog_new("Changing Uncertain Points Color");
  colorsel = GTK_COLOR_SELECTION_DIALOG(colorseldlg)->colorsel;  

  ok_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->ok_button;
  cancel_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->cancel_button;
  help_button = GTK_COLOR_SELECTION_DIALOG (colorseldlg)->help_button;
  gtk_widget_hide(cancel_button);
  gtk_widget_hide(help_button);
  gtk_signal_connect (GTK_OBJECT (ok_button), "clicked",
                      GTK_SIGNAL_FUNC (CloseWin),
                      colorseldlg);

  gtk_signal_connect(GTK_OBJECT(colorsel), "color_changed",
		     (GtkSignalFunc)UncertainPointsColorChanged, Tmp); 
  gtk_widget_show(colorseldlg);
  gdk_window_set_background (Tmp->UncertainPointsColorDrawingArea->window, 
			     &Tmp->UncertainPointsColor);
  gdk_window_clear (Tmp->UncertainPointsColorDrawingArea->window);    
}

static void drawingarea_configure(GtkWidget *drawingarea, GdkEvent *event,
				  GdkColor *color)
{
  gdk_window_set_background (drawingarea->window, color);
  gdk_window_clear (drawingarea->window);
}

void GLevelCurveProperties()
{ 
  GtkWidget *window;
  GtkWidget *vbox,*hbox,*hbox1;
  GtkWidget *Label;
  GtkWidget *ShowTriang;
  GtkWidget *eventbox;
  GtkWidget *button;
  char s[100];
  struct GLevelCurve *Tmp=Curves;
  GtkAdjustment *Prio_adj,*MinP_adj,*MaxP_adj;
  GtkTooltips *tooltips;
  int DisplayLevelCurve (GtkWidget *dummy,struct GLevelCurve *Lev);

  tooltips = gtk_tooltips_new ();

  window = gtk_window_new (GTK_WINDOW_DIALOG);
  gtk_widget_set_name (window, "window");
  gtk_object_set_data (GTK_OBJECT (window), "window", window);
  gtk_window_set_title(GTK_WINDOW(window),"Level Curve Properties"); 
 
  vbox=gtk_vbox_new(TRUE,3);
  gtk_container_add(GTK_CONTAINER(window),vbox); 
  gtk_widget_show(vbox);

  hbox=gtk_hbox_new(TRUE,3);
  gtk_container_add(GTK_CONTAINER(vbox),hbox);
  gtk_widget_show(hbox);
  
  Label=gtk_label_new("Curve");
  gtk_container_add(GTK_CONTAINER(hbox),Label);
  gtk_widget_show(Label);

  Label=gtk_label_new("cells");
  gtk_container_add(GTK_CONTAINER(hbox),Label);
  gtk_widget_show(Label);

  Label=gtk_label_new("Compute");
  gtk_container_add(GTK_CONTAINER(hbox),Label);
  gtk_widget_show(Label);

  Label=gtk_label_new("Method");
  gtk_container_add(GTK_CONTAINER(hbox),Label);
  gtk_widget_show(Label);

  Label=gtk_label_new("Priority");
  gtk_container_add(GTK_CONTAINER(hbox),Label);
  gtk_widget_show(Label);

  Label=gtk_label_new("From");
  gtk_container_add(GTK_CONTAINER(hbox),Label);
  gtk_widget_show(Label);

  Label=gtk_label_new("To");
  gtk_container_add(GTK_CONTAINER(hbox),Label);
  gtk_widget_show(Label);

  Label=gtk_label_new("Colors");
  gtk_container_add(GTK_CONTAINER(hbox),Label);
  gtk_widget_show(Label);

  Label=gtk_label_new("Values");
  gtk_container_add(GTK_CONTAINER(hbox),Label);
  gtk_widget_show(Label);

  while (Tmp)
    {
      hbox=gtk_hbox_new(TRUE,3);
      gtk_widget_show(hbox);
      gtk_container_add(GTK_CONTAINER(vbox),hbox);
      sprintf(s,"%.3e",Tmp->Sigma);
      Label=gtk_label_new(s);
      gtk_container_add(GTK_CONTAINER(hbox),Label);
      gtk_widget_show(Label);
      ShowTriang=gtk_check_button_new_with_label ("Show");
      GTK_TOGGLE_BUTTON(ShowTriang)->active= Tmp->DrawTriangles;
      gtk_signal_connect(GTK_OBJECT(ShowTriang),"clicked",Triang,Tmp);
      gtk_widget_show(ShowTriang);
      gtk_tooltips_set_tip (tooltips, ShowTriang,
			    "Toggle to show or hide triangles", NULL);
      gtk_container_add(GTK_CONTAINER(hbox),ShowTriang);
      ShowTriang=gtk_check_button_new_with_label ("Curve");
      GTK_TOGGLE_BUTTON(ShowTriang)->active= !Tmp->Limit2Triangles;
      gtk_signal_connect(GTK_OBJECT(ShowTriang),"clicked",Limit2Triang,Tmp);
      gtk_widget_show(ShowTriang);
      gtk_tooltips_set_tip (tooltips, ShowTriang,
			    "limit computations to triangles???", NULL);
      gtk_container_add(GTK_CONTAINER(hbox),ShowTriang);
      MethodName(Tmp->Method,s,100);
      Label=gtk_label_new(s);
      gtk_widget_show(Label);
      gtk_container_add(GTK_CONTAINER(hbox),Label);
      Prio_adj=(GtkAdjustment *) gtk_adjustment_new(1.0,-10.0,10.0,1.0,5.00,5.0);
      Tmp->PriorityWidget=gtk_spin_button_new(Prio_adj,1.0,0);
      sprintf(s,"%d",Tmp->Priority);
      gtk_entry_set_text(& GTK_SPIN_BUTTON(Tmp->PriorityWidget)->entry,s);
      gtk_container_add(GTK_CONTAINER(hbox),Tmp->PriorityWidget);
      gtk_widget_show(Tmp->PriorityWidget);
      gtk_tooltips_set_tip (tooltips, Tmp->PriorityWidget,
			    "Raise or lower priority", NULL);
      gtk_signal_connect((GtkObject *)(&GTK_SPIN_BUTTON(Tmp->PriorityWidget)->entry),
			 "changed",PriorityChanged,Tmp);
      MinP_adj=(GtkAdjustment *) gtk_adjustment_new(1.0,1.0,Machines,1.0,1.0,1.0);
      Tmp->MinPWidget=gtk_spin_button_new(MinP_adj,1.0,0);
      sprintf(s,"%d",Tmp->MinP);
      gtk_entry_set_text(& GTK_SPIN_BUTTON(Tmp->MinPWidget)->entry,s);
      gtk_container_add(GTK_CONTAINER(hbox),Tmp->MinPWidget);
      gtk_widget_show(Tmp->MinPWidget);
      gtk_tooltips_set_tip (tooltips, Tmp->MinPWidget,
			    "First working processor", NULL);
      gtk_signal_connect((GtkObject *)(&GTK_SPIN_BUTTON(Tmp->MinPWidget)->entry),
			 "changed",MinPChanged,Tmp);
      MaxP_adj=(GtkAdjustment *) gtk_adjustment_new(1.0,1.0,Machines,1.0,1.0,1.0);
      Tmp->MaxPWidget=gtk_spin_button_new(MaxP_adj,1.0,0);
      sprintf(s,"%d",Tmp->MaxP);
      gtk_entry_set_text(& GTK_SPIN_BUTTON(Tmp->MaxPWidget)->entry,s);
      gtk_container_add(GTK_CONTAINER(hbox),Tmp->MaxPWidget);
      gtk_widget_show(Tmp->MaxPWidget);
      gtk_tooltips_set_tip (tooltips, Tmp->MaxPWidget,
			    "Last working processor", NULL);
      gtk_signal_connect((GtkObject *)(&GTK_SPIN_BUTTON(Tmp->MaxPWidget)->entry),
			 "changed",MaxPChanged,Tmp);
      hbox1=gtk_hbox_new(TRUE,3);
      gtk_widget_show(hbox1);
      gtk_container_add(GTK_CONTAINER(hbox),hbox1);

      eventbox = gtk_event_box_new ();
      gtk_widget_set_name (eventbox, "eventbox");
      gtk_widget_ref (eventbox);
      gtk_object_set_data_full (GTK_OBJECT (window), "eventbox", eventbox,
				(GtkDestroyNotify) gtk_widget_unref);
      gtk_container_add (GTK_CONTAINER (hbox1), eventbox);
      gtk_tooltips_set_tip (tooltips, eventbox,"Level Curve Color", NULL);
      gtk_widget_show(eventbox);

      Tmp->CurveColorDrawingArea = gtk_drawing_area_new (); 
      gtk_drawing_area_size (GTK_DRAWING_AREA(Tmp->CurveColorDrawingArea), 10, 10);
      gtk_widget_set_events (Tmp->CurveColorDrawingArea, GDK_BUTTON_PRESS_MASK); 
      gtk_signal_connect (GTK_OBJECT(Tmp->CurveColorDrawingArea), "button_press_event",  
			  (GtkSignalFunc)CurveColorAreaEvent,Tmp); 
      gtk_container_add (GTK_CONTAINER(eventbox), Tmp->CurveColorDrawingArea); 
      gtk_widget_show (Tmp->CurveColorDrawingArea); 
      gtk_signal_connect(GTK_OBJECT(Tmp->CurveColorDrawingArea), "configure_event", 
 			 GTK_SIGNAL_FUNC(drawingarea_configure),&Tmp->CurveColor); 


      eventbox = gtk_event_box_new ();
      gtk_widget_set_name (eventbox, "eventbox");
      gtk_widget_ref (eventbox);
      gtk_object_set_data_full (GTK_OBJECT (window), "eventbox", eventbox,
				(GtkDestroyNotify) gtk_widget_unref);
      gtk_container_add (GTK_CONTAINER (hbox1), eventbox);
      gtk_tooltips_set_tip (tooltips, eventbox,"Triangles Color", NULL);
      gtk_widget_show(eventbox);

      Tmp->SegmentColorDrawingArea = gtk_drawing_area_new (); 
      gtk_drawing_area_size (GTK_DRAWING_AREA(Tmp->SegmentColorDrawingArea), 10, 10);
      gtk_widget_set_events (Tmp->SegmentColorDrawingArea, GDK_BUTTON_PRESS_MASK); 
      gtk_signal_connect (GTK_OBJECT(Tmp->SegmentColorDrawingArea), "button_press_event",  
			  (GtkSignalFunc)SegmentColorAreaEvent,Tmp); 
      gtk_container_add (GTK_CONTAINER(eventbox), Tmp->SegmentColorDrawingArea); 
      gtk_widget_show (Tmp->SegmentColorDrawingArea); 
      gtk_signal_connect(GTK_OBJECT(Tmp->SegmentColorDrawingArea), "configure_event", 
 			 GTK_SIGNAL_FUNC(drawingarea_configure),&Tmp->SegmentColor); 

      eventbox = gtk_event_box_new ();
      gtk_widget_set_name (eventbox, "eventbox");
      gtk_widget_ref (eventbox);
      gtk_object_set_data_full (GTK_OBJECT (window), "eventbox", eventbox,
				(GtkDestroyNotify) gtk_widget_unref);
      gtk_container_add (GTK_CONTAINER (hbox1), eventbox);
      gtk_tooltips_set_tip (tooltips, eventbox,"Internal points Color", NULL);
      gtk_widget_show(eventbox);
      Tmp->InternalPointsColorDrawingArea = gtk_drawing_area_new (); 
      gtk_drawing_area_size (GTK_DRAWING_AREA(Tmp->InternalPointsColorDrawingArea), 10, 10);
      gtk_widget_set_events (Tmp->InternalPointsColorDrawingArea, GDK_BUTTON_PRESS_MASK); 
      gtk_signal_connect (GTK_OBJECT(Tmp->InternalPointsColorDrawingArea), "button_press_event",
			  (GtkSignalFunc)InternalPointsColorAreaEvent,Tmp); 
      gtk_container_add (GTK_CONTAINER(eventbox), Tmp->InternalPointsColorDrawingArea); 
      gtk_widget_show (Tmp->InternalPointsColorDrawingArea); 
      gtk_signal_connect(GTK_OBJECT(Tmp->InternalPointsColorDrawingArea), "configure_event", 
 			 GTK_SIGNAL_FUNC(drawingarea_configure),&Tmp->InternalPointsColor); 

      eventbox = gtk_event_box_new ();
      gtk_widget_set_name (eventbox, "eventbox");
      gtk_widget_ref (eventbox);
      gtk_object_set_data_full (GTK_OBJECT (window), "eventbox", eventbox,
				(GtkDestroyNotify) gtk_widget_unref);
      gtk_container_add (GTK_CONTAINER (hbox1), eventbox);
      gtk_tooltips_set_tip (tooltips, eventbox,"External points Color", NULL);
      gtk_widget_show(eventbox);
      Tmp->ExternalPointsColorDrawingArea = gtk_drawing_area_new (); 
      gtk_drawing_area_size (GTK_DRAWING_AREA(Tmp->ExternalPointsColorDrawingArea), 10, 10);
      gtk_widget_set_events (Tmp->ExternalPointsColorDrawingArea, GDK_BUTTON_PRESS_MASK); 
      gtk_signal_connect (GTK_OBJECT(Tmp->ExternalPointsColorDrawingArea), "button_press_event",
			  (GtkSignalFunc)ExternalPointsColorAreaEvent,Tmp); 
      gtk_container_add (GTK_CONTAINER(eventbox), Tmp->ExternalPointsColorDrawingArea); 
      gtk_widget_show (Tmp->ExternalPointsColorDrawingArea); 
      gtk_signal_connect(GTK_OBJECT(Tmp->ExternalPointsColorDrawingArea), "configure_event", 
 			 GTK_SIGNAL_FUNC(drawingarea_configure),&Tmp->ExternalPointsColor); 

      eventbox = gtk_event_box_new ();
      gtk_widget_set_name (eventbox, "eventbox");
      gtk_widget_ref (eventbox);
      gtk_object_set_data_full (GTK_OBJECT (window), "eventbox", eventbox,
				(GtkDestroyNotify) gtk_widget_unref);
      gtk_container_add (GTK_CONTAINER (hbox1), eventbox);
      gtk_tooltips_set_tip (tooltips, eventbox,"Uncertain points Color", NULL);
      gtk_widget_show(eventbox);
      Tmp->UncertainPointsColorDrawingArea = gtk_drawing_area_new (); 
      gtk_drawing_area_size (GTK_DRAWING_AREA(Tmp->UncertainPointsColorDrawingArea),10, 10);
      gtk_widget_set_events (Tmp->UncertainPointsColorDrawingArea, GDK_BUTTON_PRESS_MASK); 
      gtk_signal_connect (GTK_OBJECT(Tmp->UncertainPointsColorDrawingArea), "button_press_event",
			  (GtkSignalFunc)UncertainPointsColorAreaEvent,Tmp); 
      gtk_container_add (GTK_CONTAINER(eventbox), Tmp->UncertainPointsColorDrawingArea); 
      gtk_widget_show (Tmp->UncertainPointsColorDrawingArea); 
      gtk_signal_connect(GTK_OBJECT(Tmp->UncertainPointsColorDrawingArea), "configure_event", 
 			 GTK_SIGNAL_FUNC(drawingarea_configure),&Tmp->UncertainPointsColor); 

      button=gtk_button_new_with_label ("Show");
      gtk_container_add(GTK_CONTAINER(hbox),button);
      gtk_widget_show(button);
      gtk_signal_connect (GTK_OBJECT (button), "clicked",
			  GTK_SIGNAL_FUNC (DisplayLevelCurve),
			  Tmp);
      Tmp=Tmp->Next;
    }
  gtk_object_set_data (GTK_OBJECT (window), "tooltips", tooltips);
  gtk_widget_show(window);
}




/***************************************** Eigenvalues Properties ******************************/


static void *EigenValues=NULL;
struct Point
{
  double r,i;
};
int AddEigenValue(double re,double im)
{
  struct Point z;
  GdkRectangle update_rect;
  if (EigenValues==NULL)
    EigenValues=CreateEmptyList();
  z.r=re;
  z.i=im;
  AddFirst(&z,sizeof(z),EigenValues);
  DrawEigenValue(&z);
  update_rect.x=0;
  update_rect.y=0;
  update_rect.width=GTK_WIDGET(LevelDrawingArea)->allocation.width;
  update_rect.height=GTK_WIDGET(LevelDrawingArea)->allocation.height;
  gtk_widget_draw(GTK_WIDGET(LevelDrawingArea), &update_rect);
}



int DrawEigenValue(void *p,void *ptr)
{
  gint x0,y0;
  struct Point *z=(struct Point *)p;
  x0=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.width
	     /(HRuler->upper-HRuler->lower)*(z->r-HRuler->lower));
  y0=(gint) (GTK_WIDGET(LevelDrawingArea)->allocation.height
	     *(z->i-VRuler->lower)/(VRuler->upper-VRuler->lower));
  gdk_draw_line(LevelPixmap,GTK_WIDGET(LevelDrawingArea)->style->white_gc,
		x0-3,y0,x0+3,y0);
  gdk_draw_line(LevelPixmap,GTK_WIDGET(LevelDrawingArea)->style->white_gc,
		x0,y0-3,x0,y0+3);
  return 0;
}

int DrawEigenValues()
{
  GdkRectangle update_rect;
  if (EigenValues==NULL)
    return;
  ForAllData(EigenValues,DrawEigenValue,NULL);
  update_rect.x=0;
  update_rect.y=0;
  update_rect.width=GTK_WIDGET(LevelDrawingArea)->allocation.width;
  update_rect.height=GTK_WIDGET(LevelDrawingArea)->allocation.height;
  gtk_widget_draw(GTK_WIDGET(LevelDrawingArea), &update_rect);
}


/************************************ Show Level Curve *************************************/

static int AddLine(void *data,void *list)
{
  struct Segment *Seg=(struct Segment *)data;
  GtkWidget *clist=(GtkWidget *)list;
  char text[400];
  char *buf[8];

  sprintf(&text[0],"%x",Seg->Flag);
  sprintf(&text[50],"%ld",Seg->xi);
  sprintf(&text[100],"%ld",Seg->yi);
  sprintf(&text[150],"%ld",Seg->xe);
  sprintf(&text[200],"%ld",Seg->ye);
  sprintf(&text[250],"%e",Seg->x);
  sprintf(&text[300],"%e",Seg->y);
  sprintf(&text[350],"%e",Seg->Sigma);
  buf[0]=&text[0];
  buf[1]=&text[50];
  buf[2]=&text[100];
  buf[3]=&text[150];
  buf[4]=&text[200];
  buf[5]=&text[250];
  buf[6]=&text[300];
  buf[7]=&text[350];
  
  gtk_clist_append ((GtkCList *)clist,buf);
  return 0;
}

int DisplayLevelCurve (GtkWidget *dummy,struct GLevelCurve *Lev)
{
  GtkWidget *window6;
  GtkWidget *vbox7;
  GtkWidget *table7;
  GtkWidget *label56;
  GtkWidget *label57;
  GtkWidget *label58;
  GtkWidget *label59;
  GtkWidget *label60;
  GtkWidget *label61;
  GtkWidget *label610;
  GtkWidget *label62;
  GtkWidget *label63;
  GtkWidget *label64;
  GtkWidget *label65;
  GtkWidget *label66;
  GtkWidget *label67;
  GtkWidget *label670;
  GtkWidget *scrolledwindow2;
  GtkWidget *clist1;
  GtkWidget *label48;
  GtkWidget *label49;
  GtkWidget *label50;
  GtkWidget *label51;
  GtkWidget *label52;
  GtkWidget *label53;
  GtkWidget *label54;
  GtkWidget *label55;
  char s[50];

  window6 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_name (window6, "window6");
  gtk_object_set_data (GTK_OBJECT (window6), "window6", window6);
  gtk_window_set_title (GTK_WINDOW (window6), "Level Curve");

  vbox7 = gtk_vbox_new (FALSE, 0);
  gtk_widget_set_name (vbox7, "vbox7");
  gtk_widget_ref (vbox7);
  gtk_object_set_data_full (GTK_OBJECT (window6), "vbox7", vbox7,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox7);
  gtk_container_add (GTK_CONTAINER (window6), vbox7);

  table7 = gtk_table_new (7, 2, TRUE);
  gtk_widget_set_name (table7, "table7");
  gtk_widget_ref (table7);
  gtk_object_set_data_full (GTK_OBJECT (window6), "table7", table7,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (table7);
  gtk_box_pack_start (GTK_BOX (vbox7), table7, FALSE, FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (table7), 18);
  gtk_table_set_row_spacings (GTK_TABLE (table7), 3);
  gtk_table_set_col_spacings (GTK_TABLE (table7), 8);

  label56 = gtk_label_new ("Level Curve");
  gtk_widget_set_name (label56, "label56");
  gtk_widget_ref (label56);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label56", label56,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label56);
  gtk_table_attach (GTK_TABLE (table7), label56, 0, 1, 0, 1,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  label57 = gtk_label_new ("Mesh Size");
  gtk_widget_set_name (label57, "label57");
  gtk_widget_ref (label57);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label57", label57,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label57);
  gtk_table_attach (GTK_TABLE (table7), label57, 0, 1, 1, 2,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  label58 = gtk_label_new ("Numerical Algorithm");
  gtk_widget_set_name (label58, "label58");
  gtk_widget_ref (label58);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label58", label58,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label58);
  gtk_table_attach (GTK_TABLE (table7), label58, 0, 1, 2, 3,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  label59 = gtk_label_new ("Numerical precision");
  gtk_widget_set_name (label59, "label59");
  gtk_widget_ref (label59);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label59", label59,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label59);
  gtk_table_attach (GTK_TABLE (table7), label59, 0, 1, 3, 4,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  label60 = gtk_label_new ("Maximum Iteration");
  gtk_widget_set_name (label60, "label60");
  gtk_widget_ref (label60);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label60", label60,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label60);
  gtk_table_attach (GTK_TABLE (table7), label60, 0, 1, 4, 5,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  label61 = gtk_label_new ("Bisection Precision");
  gtk_widget_set_name (label61, "label61");
  gtk_widget_ref (label61);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label61", label61,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label61);
  gtk_table_attach (GTK_TABLE (table7), label61, 0, 1, 5, 6,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  label610 = gtk_label_new ("Time");
  gtk_widget_set_name (label610, "label610");
  gtk_widget_ref (label610);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label610", label610,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label610);
  gtk_table_attach (GTK_TABLE (table7), label610, 0, 1, 6, 7,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  sprintf(s,"%e",Lev->Sigma);
  label62 = gtk_label_new (s);
  gtk_widget_set_name (label62, "label62");
  gtk_widget_ref (label62);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label62", label62,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label62);
  gtk_table_attach (GTK_TABLE (table7), label62, 1, 2, 0, 1,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  sprintf(s,"%e",Lev->Tau);
  label63 = gtk_label_new (s);
  gtk_widget_set_name (label63, "label63");
  gtk_widget_ref (label63);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label63", label63,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label63);
  gtk_table_attach (GTK_TABLE (table7), label63, 1, 2, 1, 2,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  switch (Lev->Method)
    {
    case LANCZOS:
      sprintf(s,"Lanczos");break;
    case DAVIDSON:
      sprintf(s,"Davidson");break;
    case TRACEMIN:
      sprintf(s,"Trace Minimization");break;
    case ARNOLDI:
      sprintf(s,"Arnoldi");break;
    }
  label64 = gtk_label_new (s);
  gtk_widget_set_name (label64, "label64");
  gtk_widget_ref (label64);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label64", label64,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label64);
  gtk_table_attach (GTK_TABLE (table7), label64, 1, 2, 2, 3,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  sprintf(s,"%e",Lev->SigmaPrec);
  label65 = gtk_label_new (s);
  gtk_widget_set_name (label65, "label65");
  gtk_widget_ref (label65);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label65", label65,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label65);
  gtk_table_attach (GTK_TABLE (table7), label65, 1, 2, 3, 4,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  sprintf(s,"%d",Lev->SigmaMaxIt);
  label66 = gtk_label_new (s);
  gtk_widget_set_name (label66, "label66");
  gtk_widget_ref (label66);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label66", label66,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label66);
  gtk_table_attach (GTK_TABLE (table7), label66, 1, 2, 4, 5,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  sprintf(s,"%e",Lev->SplitPrec);
  label67 = gtk_label_new (s);
  gtk_widget_set_name (label67, "label67");
  gtk_widget_ref (label67);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label67", label67,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label67);
  gtk_table_attach (GTK_TABLE (table7), label67, 1, 2, 5, 6,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  sprintf(s,"%.0fs",Lev->LastModified-Lev->CreatedTime);
  label670 = gtk_label_new (s);
  gtk_widget_set_name (label670, "label670");
  gtk_widget_ref (label670);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label670", label670,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label670);
  gtk_table_attach (GTK_TABLE (table7), label670, 1, 2, 6, 7,
                    (GtkAttachOptions) (0),
                    (GtkAttachOptions) (0), 0, 0);

  scrolledwindow2 = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_set_name (scrolledwindow2, "scrolledwindow2");
  gtk_widget_ref (scrolledwindow2);
  gtk_object_set_data_full (GTK_OBJECT (window6), "scrolledwindow2", scrolledwindow2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (scrolledwindow2);
  gtk_box_pack_start (GTK_BOX (vbox7), scrolledwindow2, TRUE, TRUE, 0);

  clist1 = gtk_clist_new (8);
  gtk_widget_set_name (clist1, "clist1");
  gtk_widget_ref (clist1);
  gtk_object_set_data_full (GTK_OBJECT (window6), "clist1", clist1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_container_add (GTK_CONTAINER (scrolledwindow2), clist1);
  gtk_clist_set_column_width (GTK_CLIST (clist1), 0, 80);
  gtk_clist_set_column_width (GTK_CLIST (clist1), 1, 80);
  gtk_clist_set_column_width (GTK_CLIST (clist1), 2, 80);
  gtk_clist_set_column_width (GTK_CLIST (clist1), 3, 80);
  gtk_clist_set_column_width (GTK_CLIST (clist1), 4, 80);
  gtk_clist_set_column_width (GTK_CLIST (clist1), 5, 80);
  gtk_clist_set_column_width (GTK_CLIST (clist1), 6, 80);
  gtk_clist_set_column_width (GTK_CLIST (clist1), 7, 80);
  gtk_clist_column_titles_show (GTK_CLIST (clist1));

  label48 = gtk_label_new ("Flags");
  gtk_widget_set_name (label48, "label48");
  gtk_widget_ref (label48);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label48", label48,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label48);
  gtk_clist_set_column_widget (GTK_CLIST (clist1), 0, label48);

  label49 = gtk_label_new ("Xi");
  gtk_widget_set_name (label49, "label49");
  gtk_widget_ref (label49);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label49", label49,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label49);
  gtk_clist_set_column_widget (GTK_CLIST (clist1), 1, label49);

  label50 = gtk_label_new ("Yi");
  gtk_widget_set_name (label50, "label50");
  gtk_widget_ref (label50);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label50", label50,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label50);
  gtk_clist_set_column_widget (GTK_CLIST (clist1), 2, label50);

  label51 = gtk_label_new ("Xe");
  gtk_widget_set_name (label51, "label51");
  gtk_widget_ref (label51);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label51", label51,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label51);
  gtk_clist_set_column_widget (GTK_CLIST (clist1), 3, label51);

  label52 = gtk_label_new ("Ye");
  gtk_widget_set_name (label52, "label52");
  gtk_widget_ref (label52);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label52", label52,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label52);
  gtk_clist_set_column_widget (GTK_CLIST (clist1), 4, label52);

  label53 = gtk_label_new ("Re(z)");
  gtk_widget_set_name (label53, "label53");
  gtk_widget_ref (label53);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label53", label53,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label53);
  gtk_clist_set_column_widget (GTK_CLIST (clist1), 5, label53);

  label54 = gtk_label_new ("Im(z)");
  gtk_widget_set_name (label54, "label54");
  gtk_widget_ref (label54);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label54", label54,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label54);
  gtk_clist_set_column_widget (GTK_CLIST (clist1), 6, label54);

  label55 = gtk_label_new ("S(z)");
  gtk_widget_set_name (label55, "label55");
  gtk_widget_ref (label55);
  gtk_object_set_data_full (GTK_OBJECT (window6), "label55", label55,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label55);
  gtk_clist_set_column_widget (GTK_CLIST (clist1), 7, label55);

  ForAllData(Lev->Orbit,AddLine,clist1);
  gtk_widget_show(clist1);
  gtk_widget_show(window6);
}

