#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/resource.h>
#include <math.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "sparse.h"
#include "mpi.h"
#include "messages.h"
#include "levelcurve.h"
#include "tasklist.h"

#include <netinet/in.h>

#define PORT 10000


struct HighPriorityTask
{
  struct MasterMsg Msg;
  int Tag;
};


void *List;
void *UserData;
struct ProcInfo *Proc;
int Slaves,Continue=1;
int sock_service;
static MPI_Datatype LM_Msg;

void keyboard(int signal)
{
  struct HighPriorityTask T;
  int p;
  char Buffer[100],*s;
  void MachState();

  p=read(sock_service,Buffer,100);
  Buffer[p]='\0';
  s=strtok(Buffer,"\n");
  while(s)
    {
      switch (s[0])
	{
	case 'S':/* Evaluer s(z) en un point donne */
	  {
	    double r,i,e;
	    long MaxIt,Meth;
	    T.Tag=Z_TAG;
	    if (sscanf(s,"S %le %le %le %ld %ld",&r,&i,&e,&MaxIt,&Meth)!=5)
	      return;
	    T.Msg.z1.r=r;
	    T.Msg.z1.i=i;
	    T.Msg.SigmaPrec=e;
	    T.Msg.SigmaMaxIt=MaxIt;
	    T.Msg.Method=Meth;
	    AddLowPriorityTask(&T,sizeof(T),List);
	    break;
	  }
	case 'V':
	  {
	    double r,i,e;
	    long nev,ncv,MaxIt;
	    T.Tag=EIGS_TAG;
	    if (sscanf(s,"V %le %le %ld %ld %ld %le",&r,&i,&nev,&ncv,&MaxIt,&e)!=6)
	      return;
	    T.Msg.z1.r=r;
	    T.Msg.z1.i=i;
	    T.Msg.z2.r=nev;
	    T.Msg.z2.i=ncv;
	    T.Msg.SigmaMaxIt=MaxIt;
	    T.Msg.SigmaPrec=e;
	    AddLowPriorityTask(&T,sizeof(T),List);
	    break;
	  }
	case 'T':
	  {
	    int id,bool;
	    if (sscanf(s,"T %d %d",&id,&bool)!=2)
	      return;
	    LimitLevelCurve2Triangles(id,bool);
	    break;
	  }
	case 'C': /* Creer une ligne de niveau */
	  {
	    Complex z;
	    double Tau,Sigma,SigmaPrec,SplitPrec;
	    long SigmaMaxIt;
	    int IsReal;
	    int Method;
	    int Slices;
	    double angle;
	    int i;

	    if (sscanf(s,"C %le %le %le %le %le %ld %le %d %d %d",
		       &z.r,&z.i,&Tau,&Sigma,&SigmaPrec,&SigmaMaxIt,
		       &SplitPrec,&IsReal,&Method,&Slices)==10)
	      {
		angle=2*3.1415927/Slices;
		for (i=1;i<=Slices;i++)
		  CreateLevelCurve(z,Tau,Sigma,SigmaPrec,SigmaMaxIt,
				   SplitPrec,IsReal,Method,3.1415927/2+i*angle);
	      }
	    break;
	  }
	case 'P': /* Imprimer une ligne de niveau */
	  {
	    int level;
	    if (sscanf(s,"P %d",&level)==1)
	      PrintLevelCurve(level);
	    break;
	  }
	case 'D': /* Effacer une ligne de niveau */
	  {
	    int level;
	    if (sscanf(s,"D %d",&level)==1)
	      DestroyLevelCurve(level);
	    break;
	  }	  
	case 'W': /* Sauvegarder une ligne de niveau */
	  {
	    int level;
	    char filename[100];
	    if (sscanf(s,"W %d %s",&level,filename)==2)
	      SaveLevelCurve(level,filename);
	    break;
	  }
	case 'Q': /* Quitter */
	  {
	    struct ProcInfo dummy;
	    close(sock_service);
	    while(GetTaskFromList(&T,sizeof(T),List));
	    for (p=0;p<Slaves;p++)
	      while (GetMasterMessage(p+1,&T.Msg,&T.Tag,&dummy));
	    Continue=0;
	    break;
	  }
	case 'A':
	  {
	    int Id;
	    if (sscanf(s,"A %d",&Id)==1)
	      SetLevelCurveAllProc(Id);
	    break;
	  }
	case 'E':
	  {
	    int Id,MinP,MaxP;
	    if (sscanf(s,"E %d %d %d",&Id,&MinP,&MaxP)==3)
	      SetLevelCurveProc(Id,MinP,MaxP);
	    break;
	  }
	case 'N':
	  {
	    int Id,Priority;
	    if (sscanf(s,"N %d %d",&Id, &Priority)==2)
	      SetLevelCurvePriority(Id,Priority);
	    break;
	  }
	}      s=strtok(NULL,"\n");
    }
  for(p=0;p<Slaves;p++)
    if (Proc[p].LevelCurve==NULL)
      {
	if(GetTaskFromList(&T,sizeof(T),List))
	  {
	    MPI_Send(&T.Msg,1,LM_Msg,p+1,T.Tag,MPI_COMM_WORLD);
	    Proc[p].LevelCurve=UserData;
	    Proc[p].Segment=NULL;
	  }
	else
	  if (GetMasterMessage(p+1,&T.Msg,&T.Tag,&Proc[p]))
	    {
	      MPI_Send(&T.Msg,1,LM_Msg,p+1,T.Tag,MPI_COMM_WORLD);
	    }
      } 
  MachState();
}

int creersock(int *port,int type)
{
  int desc;
  struct sockaddr_in nom;

  if ((desc=socket(AF_INET,type,IPPROTO_TCP))==-1)
    {
      perror("Could not create socket");
      return -1;
    }
  bzero((char *)&nom,sizeof(nom));
  nom.sin_port=htons(*port);
  nom.sin_addr.s_addr=INADDR_ANY;
  nom.sin_family=AF_INET;
  if (bind(desc,(struct sockaddr *)&nom,sizeof(nom))<0)
    {
      perror("Could not bind socket");
      return -1;
    }
  return desc;
}


void MachState()
{
  static char Oldtmp[100]="";
  char tmp[100];
  long p;

  tmp[0]='M';
  for (p=0;p<Slaves;p++)
    {
      if (Proc[p].LevelCurve==NULL)
	tmp[p+1]='-';
      else
	if (Proc[p].LevelCurve==UserData)
	  tmp[p+1]='u';
	else
	  tmp[p+1]='A'-1+LevelID(Proc[p].LevelCurve);
    }
  tmp[p+1]='\n';
  tmp[p+2]='\0';
  if (strcmp(tmp,Oldtmp)!=0)
    {
      strcpy(Oldtmp,tmp);
      write(sock_service,tmp,strlen(tmp));
    }
}

void MainMaster(int Size,MPI_Datatype M_Msg,MPI_Datatype S_Msg)
{
  struct SlaveMsg In;
  MPI_Status Status;
  void *A;
  char Filename[100]="matrix.mtx";
  int p,i,flag;
  struct HighPriorityTask T;
  void BroadCastMatrix(void *ParA,int Size);

  /* Creating Socket */
  int sock_ecoute;
  struct sockaddr_in adr;
  int lgadr=sizeof(adr);
  int port=PORT;

  /* Signal Data */
  sigset_t set;

  /* Creation du socket d'ecoute */

  if ((sock_ecoute=creersock(&port,SOCK_STREAM))==-1)
    {
      return;
    }
  /* Creation de la file pour les demandes de connexions */
  if (listen(sock_ecoute,1)<0)
    {
      perror("listen");
      return;
    }
  /* Attente d'une demande de connexion */
  printf("Waiting for FrontEnd...\n");
  sock_service=accept(sock_ecoute,&adr,&lgadr);
  printf("Connection established...\n");
  /*Fermeture du socket d'ecoute */
  close(sock_ecoute);
  /* Setting signal SIGIO handle */
  signal(SIGIO,keyboard);
  fcntl(sock_service,F_SETFL,O_ASYNC);
  fcntl(sock_service,F_SETOWN,getpid());
  SetSocketHandle(sock_service);
  UserData=(void *) malloc(1);
  List=CreateTaskList(50);
  Slaves=Size-1;
  Proc=(struct ProcInfo*)malloc(Slaves*sizeof(struct ProcInfo));
  memset(Proc,0,Slaves*sizeof(struct ProcInfo));
  LM_Msg=M_Msg;
  /*  Selecting Matrix ...*/

  A=SparseLoadMTX(Filename);
  if (A==NULL)
    {
      printf("Cannot open file...%s\n",Filename);
      return;
    }
   /* Step 1- Broadcasting the matrix */
  BroadCastMatrix(A,Size);
  SparseDestroy(A);
  A=NULL;
  sigemptyset(&set);
  sigaddset(&set,SIGIO);
  while(1)
    {
      /* Chercher une machine disponible */
      i=0;
      for(p=0;p<Slaves;p++)
	if (Proc[p].LevelCurve==NULL)
	  {
	    /* Bloquer la signalisation pour assurer l'exclusion mutuelle dans la liste des taches */
	    sigprocmask(SIG_BLOCK,&set,NULL);
	    /* Tester si il y a d'autres messages plus prioritaires */
	    flag=GetTaskFromList(&T,sizeof(T),List);
	    /* Debloquer le signal SIG_IO */
	    sigprocmask(SIG_UNBLOCK,&set,NULL);
	    if(flag)
	      {
		i++;
		MPI_Send(&T.Msg,1,M_Msg,p+1,T.Tag,MPI_COMM_WORLD);
		Proc[p].LevelCurve=UserData;
	      }
	    else
	      if (GetMasterMessage(p+1,&T.Msg,&T.Tag,&Proc[p]))
		{
		  i++;
		  MPI_Send(&T.Msg,1,M_Msg,p+1,T.Tag,MPI_COMM_WORLD);
		}
	  }
	else
	  i++;
      MachState();
      if (i==0 && Continue==0)
 	break;
      if (i)
	MPI_Recv(&In,1,S_Msg,MPI_ANY_SOURCE,MPI_ANY_TAG,MPI_COMM_WORLD,&Status);
      else
	{
	  pause();
	  continue;
	}
      if (Continue)
	{
	  if (Proc[Status.MPI_SOURCE-1].LevelCurve==UserData)
	    {
	      if (Status.MPI_TAG==Z_TAG)
		{
		  sprintf(Filename,"B %e %e %e %ld\n",In.z.r,In.z.i,In.Sigma,In.It);
		  write(sock_service,Filename,strlen(Filename));
		}
	      if (Status.MPI_TAG==EIGS_TAG)
		{
		  if (In.Status==1)
		    {
		      sprintf(Filename,"V %e %e\n",In.z.r,In.z.i);
		      write(sock_service,Filename,strlen(Filename));
		      sprintf(Filename,"V %e %e\n",In.ze.r,In.ze.i);
		      write(sock_service,Filename,strlen(Filename));
		      while (In.Status==1)
			{
			  MPI_Recv(&In,1,S_Msg,
				   Status.MPI_SOURCE,EIGS_TAG,MPI_COMM_WORLD,
				   &Status);
			  if (In.Status==2)
			    {
			      sprintf(Filename,"V %e %e\n",In.z.r,In.z.i);
			      write(sock_service,Filename,strlen(Filename));
			      sprintf(Filename,"V %e %e\n",In.ze.r,In.ze.i);
			      write(sock_service,Filename,strlen(Filename));
			    }
			  else
			    {
			      sprintf(Filename,"V %e %e\n",In.z.r,In.z.i);
			      write(sock_service,Filename,strlen(Filename));
			    }
			}
		    }
		  else
		    {
		      sprintf(Filename,"X ARPACK: Maximum Iteration Reached...");
		      write(sock_service,Filename,strlen(Filename));
		    }
		}
	    }
	  else
	    {
	    /* Bloquer la signalisation pour assurer l'exclusion mutuelle dans la liste des taches */
	    sigprocmask(SIG_BLOCK,&set,NULL);
	    /* Analyser le message et creer les taches necessaires */
	    ProcessSlaveMessage(Status.MPI_TAG,&In,&Proc[Status.MPI_SOURCE-1]);
	    /* Debloquer le signal SIG_IO */
	    sigprocmask(SIG_UNBLOCK,&set,NULL);
	    }
	}
      Proc[Status.MPI_SOURCE-1].LevelCurve=NULL;
    }
  free(UserData);
  close(sock_service);
}

