#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/resource.h>
#include <math.h>
#include "sparse.h"
#include "mpi.h"
#include "messages.h"
#include "tasklist.h"
#include "sigmamin.h"
#include "eigs.h"
#include "split.h"

#define COSPI_3   0.5
#define SINPI_3   0.866025404


static int Start(void *A,
         Complex z,
         double LevelCurve,
         double Tau,
         struct SigmaMinParam *p,
         double angle,
         long *xi,long *yi,
         long *xe,long *ye)
{
  double alpha=Tau;
  char Status;
  double s;
  long MaxIt=p->SigmaMaxIt;
  double re,im;

  Complex zi,ze,zl;
  long ki,li,ke,le;
  double min,max;
  double e,e1,e2;
  long i;

  /* Trouver un sommet du triangle contenant z verifiant SigmaMin(sommet) < LevelC
urve */
  ki=floor(z.i/Tau/SINPI_3);
  li=floor(z.r/Tau-COSPI_3*ki);
  zi.r=li*Tau+ki*Tau*COSPI_3;
  zi.i=ki*Tau*SINPI_3;
  if (SigmaMin(A,zi,p)>LevelCurve)
    {
      li++;
      zi.r=li*Tau+ki*Tau*COSPI_3;
      p->SigmaMaxIt=MaxIt;
      if (SigmaMin(A,zi,p)>LevelCurve)
    {
      ki++;
      zi.r=li*Tau+ki*Tau*COSPI_3;
      zi.i=ki*Tau*SINPI_3;
      p->SigmaMaxIt=MaxIt;
      if (SigmaMin(A,zi,p)>LevelCurve)
        {
          li--;
          zi.r=li*Tau+ki*Tau*COSPI_3;
          zi.i=ki*Tau*SINPI_3;
          p->SigmaMaxIt=MaxIt;
          if (SigmaMin(A,zi,p)>LevelCurve)
        return 0;
        }          
    }
    }
  re=cos(angle);
  im=sin(angle);
  while (1)
    {
      zl.r=zi.r+alpha*re;
      zl.i=zi.i+alpha*im;
      alpha=2*alpha;
      p->SigmaMaxIt=MaxIt;
      s=SigmaMin(A,zl,p);
      if (s>LevelCurve)
    {
      /* Trouver un sommet du triangle contenant zl verifiant SigmaMin(sommet)>Lev
elCurve */
      ke=floor(zl.i/Tau/SINPI_3);
      le=floor(zl.r/Tau-COSPI_3*ke);
      ze.r=le*Tau+ke*Tau*COSPI_3;
      ze.i=ke*Tau*SINPI_3;
      if (SigmaMin(A,ze,p)<LevelCurve)
        {
          le++;
          ze.r=le*Tau+ke*Tau*COSPI_3;
          p->SigmaMaxIt=MaxIt;
          if (SigmaMin(A,ze,p)<LevelCurve)
        {
          ke++;
          ze.r=Tau*(le+ke*COSPI_3);
          ze.i=ke*Tau*SINPI_3;
          p->SigmaMaxIt=MaxIt;
          if (SigmaMin(A,ze,p)<LevelCurve)
            {
              le--;     
              ze.r=le*Tau+ke*Tau*COSPI_3;
              ze.i=ke*Tau*SINPI_3;
              p->SigmaMaxIt=MaxIt;
              if (SigmaMin(A,ze,p)<LevelCurve)
            continue;
            }
        }
        }
      break;
    }
    }
  /* Tester le point 0.5 */
  z.r=zi.r+(le-li)*Tau;
  z.i=zi.i;
  p->SigmaMaxIt=MaxIt;
  if (SigmaMin(A,z,p)<LevelCurve)
    {
      /* le segment correspond a [0.5,1] */
      if (ki<ke)
    {
      for(i=ki+1;i<=ke;i++)
        {
          z.r=zi.r+Tau*(le-li+(i-ki)*COSPI_3);
          z.r=zi.i+Tau*((i-ki)*SINPI_3);  
          p->SigmaMaxIt=MaxIt;
          if (SigmaMin(A,z,p)>LevelCurve)
        {
          *xi=*xe=le;
          *yi=i-1;
          *ye=i;
          break;
        }
        }
    }
      else
    {
      for(i=ki-1;i>=ke;i--)
        {
          z.r=zi.r+Tau*(le-li+(i-ki)*COSPI_3);
          z.r=zi.i+Tau*((i-ki)*SINPI_3);
          p->SigmaMaxIt=MaxIt;
          if (SigmaMin(A,z,p)>LevelCurve)
        {
          *xi=*xe=le;
          *yi=i+1;
          *ye=i;
          break;
        }
        }
    }
    }
  else
    {
      /* Le segment correspond a [0,0.5] */
      if (le>li)
    {
      for(i=li+1;i<=le;i++)
        {
          z.r=zi.r+(i-li)*Tau;
          z.i=zi.i;
          p->SigmaMaxIt=MaxIt;
          if (SigmaMin(A,z,p)>LevelCurve)
        {
          *xi=i-1;
          *xe=i;
          *yi=*ye=ki;
          break;
        }
        }
    }        
      else
    {
      for(i=li-1;i>=le;i--)
        {
          z.r=zi.r+(i-li)*Tau;
          z.i=zi.i;
          p->SigmaMaxIt=MaxIt;
          if (SigmaMin(A,z,p)>LevelCurve)
        {
          *xi=i+1;
          *xe=i;
          *yi=*ye=ki;
          break;
        }
        }
    }

    }
  return 1;
}     


static int LocateLevelCurve(void *A,
			    Complex z,
			    double LevelCurve,
			    Complex *zl,
			    double Tau,
			    double SplitPrecision,
			    struct SigmaMinParam *p,
			    double angle)
{
  double alpha=Tau;
  char Status;
  double s;
  long MaxIt=p->SigmaMaxIt;
  double re,im;

  if (SigmaMin(A,z,p)>LevelCurve)
    return 0;
  re=cos(angle);
  im=sin(angle);
  while (1)
    {
      zl->r=z.r+alpha*re;
      zl->i=z.i+alpha*im;
      alpha=2*alpha;
      p->SigmaMaxIt=MaxIt;
      s=SigmaMin(A,*zl,p);
      if (s>LevelCurve)
	break;
    }
  *zl=Split(z,*zl,A,LevelCurve,p,SplitPrecision,&alpha,&Status);
  if (Status)
    return -1;
  return 1;
}


static int FindFirstSegment(void *A,
			    Complex zl,double Tau,
			    double Epsilon,
			    struct SigmaMinParam *p,
			    long *xi,long *yi,
			    long *xe,long *ye,
			    double *Si,double *Se)
{
  double k,l;
  Complex z1,z2,z3,z4;
  double s1,s2,s3,s4;
  long MaxIt=p->SigmaMaxIt;

  k=zl.i/Tau/SINPI_3;
  l=zl.r/Tau-COSPI_3*k;
  l=floor(l);
  k=floor(k);
  z1.r=l*Tau+k*Tau*COSPI_3;
  z1.i=k*Tau*SINPI_3;
  p->SigmaMaxIt=MaxIt;
  s1=SigmaMin(A,z1,p);
  z2.r=z1.r+Tau;
  z2.i=z1.i;
  p->SigmaMaxIt=MaxIt;
  s2=SigmaMin(A,z2,p);
  if ((s1-Epsilon)*(s2-Epsilon)<0)
    {
      if (s1<Epsilon)
	{
	  *xi=l;*yi=k;
	  *xe=l+1;*ye=k;
	  *Si=s1;*Se=s2;
	}
      else
	{
	  *xe=l;*ye=k;
	  *xi=l+1;*yi=k;
	  *Si=s2;*Se=s1;
	}
      return 1;
    }
  z3.r=z1.r+Tau*COSPI_3;
  z3.i=z1.i+Tau*SINPI_3;
  p->SigmaMaxIt=MaxIt;
  s3=SigmaMin(A,z3,p);
  if ((s1-Epsilon)*(s3-Epsilon)<0)
    {
      if (s1<Epsilon)
	{
	  *xi=l;*yi=k;
	  *xe=l;*ye=k+1;
	  *Si=s1;*Se=s3;
	}
       else
	{
	  *xe=l;*ye=k;
	  *xi=l;*yi=k+1;
	  *Si=s3;*Se=s1;
	}
      return 1;
   }
  if ((s2-Epsilon)*(s3-Epsilon)<0)
    {
      if (s2<Epsilon)
	{
	  *xi=l+1;*yi=k;
	  *xe=l;*ye=k+1;
	  *Si=s2;*Se=s3;
	}
       else
	{
	  *xe=l+1;*ye=k;
	  *xi=l;*yi=k+1;
	  *Si=s3;*Se=s2;
	}
      return 1;
   }
  z4.r=z3.r+Tau;
  z4.i=z3.i;
  p->SigmaMaxIt=MaxIt;
  s4=SigmaMin(A,z4,p);
  if ((s4-Epsilon)*(s2-Epsilon)<0)
    {
      if (s2<Epsilon)
	{
	  *xi=l+1;*yi=k;
	  *xe=l+1;*ye=k+1;
	  *Si=s2;*Se=s4;
	}
       else
	{
	  *xe=l+1;*ye=k;
	  *xi=l+1;*yi=k+1;
	  *Si=s4;*Se=s2;
	}
      return 1;
   }
  if ((s4-Epsilon)*(s3-Epsilon)<0)
    {
      if (s3<Epsilon)
	{
	  *xi=l;*yi=k+1;
	  *xe=l+1;*ye=k+1;
	  *Si=s3;*Se=s4;
	}
       else
	{
	  *xe=l;*ye=k;
	  *xi=l+1;*yi=k+1;
	  *Si=s4;*Se=s3;
	}
      return 1;
   }
  return 0;
}


void MainSlave(int Rank,MPI_Datatype M_Msg,MPI_Datatype S_Msg)
{
  void *A=NULL;
  MPI_Status Status;
  struct MasterMsg In;
  struct SlaveMsg Out;
  struct SigmaMinParam p;
  void *RecvMatrix();
  Complex zl,*v;
  long xi,yi,xe,ye;
  long nev,ncv,i;

  A=RecvMatrix();
  while(1)
    {
      MPI_Recv(&In,1,M_Msg,0,MPI_ANY_TAG,MPI_COMM_WORLD,&Status);
      switch(Status.MPI_TAG)
	{
	case ENDPROG_TAG:
          printf("[%d] Received ENDPROG...Exiting\n",Rank);
	  fflush(stdout);
          SparseDestroy(A);
          return;
	case INTERVAL_TAG:
	  p.Method=In.Method;
	  p.SigmaPrec=In.SigmaPrec;
	  p.SigmaMaxIt=In.SigmaMaxIt;
	  Out.z=Split(In.z1,In.z2,A,In.Sigma,&p,In.SplitPrec,&Out.Sigma,&Out.Status);
	  MPI_Send(&Out,1,S_Msg,0,INTERVAL_TAG,MPI_COMM_WORLD);
          break;
	case Z_TAG:
	  p.Method=In.Method;
	  p.SigmaPrec=In.SigmaPrec;
	  p.SigmaMaxIt=In.SigmaMaxIt;
	  if (p.Method==-1)
	    {
	      Out.Sigma=In.Sigma;
	      Out.Status=2;
	    }
	  else
	    {
	      Out.Sigma=SigmaMin(A,In.z1,&p);
	      Out.Status=(p.SigmaMaxIt<In.SigmaMaxIt);
	    }
	  Out.z=In.z1;
	  Out.It=p.SigmaMaxIt;
	  MPI_Send(&Out,1,S_Msg,0,Z_TAG,MPI_COMM_WORLD);
	  break;
	case Z1_TAG:
	  p.Method=In.Method;
	  p.SigmaPrec=In.SigmaPrec;
	  p.SigmaMaxIt=In.SigmaMaxIt;
	  if (p.Method==-1)
	    {
	      Out.Sigma=In.Sigma;
	      Out.Status=2;
	    }
	  else
	    {
	      Out.Sigma=SigmaMin(A,In.z1,&p);
	      Out.Status=(p.SigmaMaxIt<In.SigmaMaxIt);
	    }
	  Out.z=In.z1;
	  Out.It=p.SigmaMaxIt;
	  MPI_Send(&Out,1,S_Msg,0,Z1_TAG,MPI_COMM_WORLD);
	  break;
	case STARTUP_TAG:
	  p.Method=In.Method;
	  p.SigmaPrec=In.SigmaPrec;
	  p.SigmaMaxIt=In.SigmaMaxIt;
	  Out.Status=0;
	  /* Premier demarrage */
	  if (LocateLevelCurve(A,In.z1,In.Sigma,&zl,In.z2.r,1e-3,&p,In.z2.i)
	      &&
	      FindFirstSegment(A,zl,In.z2.r,In.Sigma,&p,&xi,&yi,&xe,&ye,&Out.ze.r,&Out.ze.i))
	    {
	      Out.Status=1;
	      Out.xi=xi;
	      Out.yi=yi;
	      Out.xe=xe;
	      Out.ye=ye;
	    }
	  /*	  else 
	    if (Start(A,In.z1,In.Sigma,In.z2.r,&p,In.z2.i,&xi,&yi,&xe,&ye)==1)
	      {
		Out.Status=1;
		Out.z.r=xi;
		Out.z.i=yi;
		Out.ze.r=xe;
		Out.ze.i=ye;
		}*/
	  MPI_Send(&Out,1,S_Msg,0,STARTUP_TAG,MPI_COMM_WORLD);
	  break;
	case EIGS_TAG:
	  nev=(long) In.z2.r;
	  ncv=(long) In.z2.i;
	  if (ncv<nev+1)
	    ncv=nev+1;
	  v=(Complex *)malloc((nev+1)*sizeof(Complex));
	  if (Eigs(A,In.z1,nev,ncv,In.SigmaMaxIt,In.SigmaPrec,v)==1)
	    {
	      Out.Status=1;
	      for(i=0;i<nev/2;i++)
		{
		  Out.z=v[2*i];
		  Out.ze=v[2*i+1];
		  MPI_Send(&Out,1,S_Msg,0,EIGS_TAG,MPI_COMM_WORLD);
		}
	      if (nev%2)
		{
		  Out.z=v[nev-1];
		  Out.Status=2;
		}
	      else
		Out.Status=3;
	      MPI_Send(&Out,1,S_Msg,0,EIGS_TAG,MPI_COMM_WORLD);		  
		
	    }
	  else
	    {
	      Out.Status=0;
	      MPI_Send(&Out,1,S_Msg,0,EIGS_TAG,MPI_COMM_WORLD);
	    }
	  free(v);
	  break;
	}
    }
}
