#include "sigmamin.h"
#include "sparse.h"
#include <stdio.h>
#include <math.h>

static doublereal *GVect=NULL;
static Complex *V0;
static long DimVect;

extern int dstebz_(char *range,char *order,
		   integer *n,
		   doublereal *vl,doublereal *vu,
		   integer *il,integer *iu,
		   doublereal *abstol,doublereal *d__,doublereal *e,
		   integer *m,integer *nsplit,
		   doublereal *w,
		   integer *iblock,integer *isplit,
		   doublereal *work,
		   integer *iwork,integer *info,
		   ftnlen range_len,
		   ftnlen order_len);
extern int dstein_(integer *n,
		   doublereal *d__,doublereal *e,
		   integer *m,
		   doublereal *w,
		   integer *iblock,integer *isplit,
		   doublereal *z__,
		   integer *ldz,
		   doublereal *work,
		   integer *iwork,integer *ifail,integer *info);

static double SigmaMinLanczos(void *ExtA,Complex z,
			      double eps,long *MaxIt)
{
  extern void zdscal_(integer *,double *,Complex *,integer *);
  extern double dznrm2_(integer *n,Complex *v,integer *inc);
  extern double dnrm2_(integer *n,double *v,integer *inc);
  extern void zaxpy_(integer *,Complex *,Complex *,integer *,Complex *,integer *);
  extern void zdotc_(Complex *,integer *,Complex *,integer *,Complex *,integer *);
  Complex *vj,*vj_1,*wj,*tmp,*buffer; /* Working Space */
  long _2n;
  Complex alpha;
  double beta;
  long inc=1,j; /* Increment of vector components */
  double bestsigma;
  long max=*MaxIt;

  /* dstebz_ parameters  */
  integer j1,m,nsplit,*isplit,*iblock,*iWork,*iFail,info,Un=1,i;
  double DZero=0.0,*w,*Work,*d,*e,*Vect;

  static integer statistics=0;
  /* QR decomposition parameters */
  void *A;
  long *P;
  void *R;
  if (max>=SparseColumns(ExtA))
    max=SparseColumns(ExtA)-1;
  if (max<10)
    max=10;
  _2n=2*SparseColumns(ExtA);
  z.i=-z.i;
  z.r=-z.r;
  A=SparseAdd_zI(ExtA,z); 
  R=SparseMFQR(A);
  SparseDestroy(A);
  vj_1=buffer=(Complex *)malloc(3*_2n*sizeof(Complex));
  wj=&vj_1[_2n];
  vj=&wj[_2n];
  /* Initialize vj */
  if (GVect==NULL)
    {
      GVect=(doublereal *)malloc(4*_2n*sizeof(doublereal));
      memset(GVect,0,4*_2n*sizeof(doublereal));
      V0=(Complex *)malloc(_2n*sizeof(Complex));
      memset(V0,0,_2n*sizeof(Complex));
    }
  for(j=0;j<_2n;j++) 
    {
      vj[j].r=1.0+1e8*V0[j].r;
      vj[j].i=1.0+1e8*V0[j].i;
    } 
  beta=dznrm2_(&_2n,vj,&inc);
  beta=1.0/beta;
  zdscal_(&_2n,&beta,vj,&inc); /* || vj ||=1 */ 
  memset(vj_1,0,_2n*sizeof(Complex)); /* vj_1=0 */
  beta=0.0;
  d=(double *)malloc((12*max)*sizeof(double)); /* max */
  e=d+max;   /* max */
  w=e+max;   /* max */
  Work=w+max;  /* 5 max */
  Vect=Work+4*max;  /* 4max */
  iblock=(integer *)malloc(5*max*sizeof(integer)); /* max */
  isplit=iblock+max;  /* max */
  iWork=isplit+max;   /* 3max */
  iFail=iWork+max;
  memset(V0,0,_2n*sizeof(Complex));
  for(j=0;j<max;)
  {
    MatProd(R,vj,wj);      /* wj=A*vj */
    alpha.r=-beta;
    alpha.i=0.0;
    zaxpy_(&_2n,&alpha,vj_1,&inc,wj,&inc);  /* wj=wj-beta*vj_1 */ 
    zdotc_(&alpha,&_2n,wj,&inc,vj,&inc); /* apha=wj'*vj */
    d[j]=alpha.r;
    alpha.r=-alpha.r;
    alpha.i=0.0;
    zaxpy_(&_2n,&alpha,vj,&inc,wj,&inc);  /* wj=wj-alpha*vj */
    beta=dznrm2_(&_2n,wj,&inc);  /* beta=|wj| */
    e[j]=beta;
    alpha.r=1/beta;
    zdscal_(&_2n,&alpha.r,wj,&inc); /* wj=wj/beta */
    /* Calcul du vecteur de depart pour l'appel suivant */
    if (j<DimVect)
      {
	alpha.r=GVect[j]+2*GVect[j+DimVect]
	  +4*GVect[j+2*DimVect]+8*GVect[j+3*DimVect];
	alpha.i=0;
	zaxpy_(&_2n,&alpha,vj,&inc,V0,&inc);
      }
    
    /* Reactualiser les indices */
    tmp=vj_1;
    vj_1=vj;
    vj=wj;
    wj=tmp;
    j++;
    if ((j>5 && j% 5 ==0) || j==max)
    {
      /* WARNING: IF MORE EIGENVALUES ARE REQUESTED, WE SHOULD INCREASE THE SIZE OF VECT */
      j1=j-4; 
      /* Use Sturm iteration to locate largest 2 eigenvalues */
      dstebz_("I","B",&j,NULL,NULL,&j1,&j,&DZero,
	      d,e,&m,&nsplit,w,iblock,isplit,Work,iWork,&info,Un,Un);
      if (info)
	{
	  printf("Sturm iteration did not converge Info=%ld\n",info);
	  continue;
	}
      j1=4;
      /* Use inverse iteration to compute eigenvectors */
      dstein_(&j,d,e,&j1,w,iblock,isplit,Vect,&j,Work,iWork,iFail,&info);
      i=0;
      for (j1=1;j1<4;j1++)
	if (w[j1]>w[i])
	  i=j1;
      bestsigma=1/w[i];
      alpha.r=beta*Vect[i*j+j-1]/w[i] ;
      if (alpha.r<0)
	alpha.r=-alpha.r;
      if (alpha.r<eps)
	{
	  bestsigma=1/w[i];
	  break;
	}
      
    }
  }
  memcpy(GVect,Vect,4*j*sizeof(doublereal));
  DimVect=j;
  SparseDestroy(R);
  free(iblock);
  free(d);
  free(buffer);
  *MaxIt=j;
 /* statistics+=j;
  printf("SigmaMin=%ld, statistics=%ld\n",j,statistics);fflush(stdout);*/
  return bestsigma;
}





/* static double SigmaMinTrace(void *ExtA,Complex z) */
/* { */
/*   long n,i,k,inc=1; */
/*   double f; */
/*   Complex mcz,mz,d,a; */
/*   Complex *V,*T,*R,*W,*B,*Y; */

/*   extern double dznrm2_(); */
/*   n=2*A->Columns; */
/*   V=(Complex *)malloc(6*n*sizeof(Complex)); */
/*   T=&V[n]; */
/*   R=&V[2*n]; */
/*   W=&V[3*n]; */
/*   B=&V[4*n]; */
/*   Y=&V[5*n]; */
/*   for(i=0;i<n;i++) */
/*     V[i].r=V[i].i=1.0; */
/*   f=1/dznrm2_(&n,V,&inc); */
/*   zdscal_(&n,&f,V,&inc); */
/*   mz.r=-z.r; */
/*   mz.i=-z.i; */
/*   mcz.r=-z.r; */
/*   mcz.i=z.i; */
/*   for(k=0;k<SigmaMaxIt;k++) */
/*     { */
/*       SparseVect(A,'N',&V[A->Columns],W,NULL); */
/*       zaxpy_(&n,&mz,&V[A->Columns],&inc,Y,&inc); */
/*       SparseVect(A,'H',T,W,NULL); */
/*       zaxpy_(&n,&mcz,T,&inc,W,&inc); */
/*       zdotc_(&d,&n,V,&inc,W,&inc);  */
/*       d.r=-d.r; */
/*       d.i=0.0; */
/*       SparseVect(A,'H',T,R,NULL); */
/*       zaxpy_(&n,&d,V,&inc,R,&inc); */
/*       f=dznrm2_(&n,R,&inc); */
/*       printf("Trace=%f, s=%f\n",f,sqrt(-d.r)); */
/*       fflush(stdout); */
/*       if (f<SigmaPrec) */
/*         { */
/* 	  free(V); */
/* 	  return sqrt(-d.r); */
/* 	} */
/*       gc(A,z,V,GCMaxIt,GCPrec,B); */
/*       zdotc_(&a,&n,V,&inc,B,&inc); */
/*       f=a.r*a.r+a.i*a.i; */
/*       a.r=a.r/f; */
/*       a.i=-a.i/f; */
/*       memset(V,0,n*sizeof(Complex)); */
/*       zaxpy_(&n,&a,B,&inc,V,&inc); */
/*       f=1/dznrm2_(&n,V,&inc); */
/*       zdscal_(&n,&f,V,&inc); */
/*     }  */
/*   free(V); */
/*   return sqrt(-d.r); */

/* } */

double SigmaMinArnoldi(void *ExtA,Complex z,long nev,long ncv,
		       long *pmaxitr,double tol)
{
  long maxitr=*pmaxitr,it;
  static char bmat[]="I";
  static integer mode, info;
  static logical rvec;
  static integer ierr;
  static doublecomplex *d__;
  static integer  n;
  static doublecomplex *v, sigma;
  static char which[]="LM";
  static doublecomplex *resid;
  static doublecomplex *workd;
  static integer ipntr[14];
  static doublecomplex *workl,*w;
  static doublereal *rwork;
  static integer iparam[11];
  static logical *select;
  static integer ishfts;
  extern int znaupd_();
  extern int zneupd_();
  static integer lworkl;
  static doublecomplex *workev;
  static integer ido;
  void *A,*R;
  long *P;
  double ret;

  z.i=-z.i;
  z.r=-z.r;
  n=2*SparseColumns(ExtA);
  A=SparseAdd_zI(ExtA,z); 
  R=SparseMFQR(A);
  SparseDestroy(A);


  resid=(doublecomplex*)malloc(n*sizeof(doublecomplex));
  w=(doublecomplex*)malloc(n*sizeof(doublecomplex));
  v=(doublecomplex*)malloc(n*ncv*sizeof(doublecomplex));
  workd=(doublecomplex*)malloc(3*n*sizeof(doublecomplex));
  lworkl=3*ncv*ncv+5*ncv;
  workl=(doublecomplex*)malloc(lworkl*sizeof(doublecomplex));
  rwork=(doublereal*)malloc(ncv*sizeof(doublereal));
  workev=(doublecomplex *)malloc(2*ncv*sizeof(doublecomplex));
  select=(logical *)malloc(ncv*sizeof(logical));
  lworkl = ncv * ncv * 3 + ncv * 5;
  d__=(doublecomplex *)malloc((nev+1)*sizeof(doublecomplex));
  ido = 0;
  info = 0;
  
  ishfts = 1;
  mode = 1;
  
  iparam[0] = ishfts;
  iparam[2] = maxitr;
  iparam[6] = mode;
  rvec = TRUE_;
  it=0;
  while(1)
    {
      it++;
      znaupd_(&ido, bmat, &n, which, &nev, &tol, resid, 
	      &ncv, v, &n, iparam, ipntr, workd, workl, 
	      &lworkl, rwork, &info);
      if (ido == -1 || ido == 1) 
	{
	  MatProd(R,&workd[ipntr[0] - 1],&workd[ipntr[1]-1]);
	  continue;
	}
      if (info < 0) 
	{
	  SparseDestroy(R);
	  free(resid);
	  free(w);
	  free(v);
	  free(workd);
	  free(workl);
	  free(rwork);
	  free(workev);
	  free(select);
	  free(d__);
	  printf("Info=%ld\n",info);
	  return -1.0;
	}
      else 
	{
	  zneupd_(&rvec, "A", select, d__, v, &n, &sigma, 
		  workev, bmat, &n,which, &nev, &tol, resid, &ncv,
		  v, &n, iparam, ipntr, workd, workl, &lworkl, 
		  rwork, &ierr);
	  if (ierr != 0)
	    {
	      SparseDestroy(R);
	      free(resid);
	      free(w);
	      free(v);
	      free(workd);
	      free(workl);
	      free(rwork);
	      free(workev);
	      free(select);
	      free(d__);
	      return -1.0;
	    }
	}
      break;
    }
  SparseDestroy(R);
  free(resid);
  free(w);
  free(v);
  free(workd);
  free(workl);
  free(rwork);
  free(workev);
  free(select);
  ret=d__[0].r>0?1/d__[0].r:-1/d__[0].r;
  free(d__);
  if (info == 1) 
    {
      printf(" Maximum number of iterations reached.\n");
      return -1;
    }
  if (info == 3) 
    {
      printf(" No shifts could be applied during implicit Arnoldi\n");
      printf("update, try increasing NCV.");
      return -1;
    }
  *pmaxitr=it;
  return ret;
}


double SigmaMin(void *A,Complex z,struct SigmaMinParam *p)
{
  long n;
  switch (p->Method)
    {
    case LANCZOS:
      return SigmaMinLanczos(A,z,p->SigmaPrec,&p->SigmaMaxIt);
    case TRACEMIN:
    case DAVIDSON:
      printf("Methods not yet implemented...\n");
      return -1;
    case ARNOLDI:
      n=SparseColumns(A);
      return SigmaMinArnoldi(A,z,3,50>n?n:50,&p->SigmaMaxIt,p->SigmaPrec);
    }
  return -1;
}
